#! /usr/bin/env python

"""File:        generator.py
   Author:      Nathan Robinson
   Contact:     nathan.m.robinson@gmail.com
   Date:        2013-11-13

   Desctiption: Generate cave diver PDDL problems.

   Lisence:     Copyright (c) Year 2013, Nathan Robinson <nathan.m.robinson@gmail.com>
                                         Christian Muise <christian.muise@gmail.com>
                                         Charles Gretton <charles.gretto@gmail.com>

                Permission to use, copy, modify, and/or distribute this software for any
                purpose with or without fee is hereby granted, provided that the above
                copyright notice and this permission notice appear in all copies.

                THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
                WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
                MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
                SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
                WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
                ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR
                IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

            To do:
            1. partial order reduction option
            2. Make it easier to specify what tunnels to make
            3. document and send in
"""

import itertools
import random

import generator_cmd_line

eps = 0.01


def make_caves(branch_depths):
    """Generate a random tree with the branches of the given lengths.
    Return the nodes, the node depths, and the leaf nodes.

    ([int]) -> [(int, int)], [int], [int]
    """
    edges = [(x, x + 1) for x in range(branch_depths[0])]
    node_depths = list(range(branch_depths[0] + 1))
    nodes = branch_depths[0] + 1
    leaves = [nodes - 1]
    for branch in branch_depths[1:]:
        junction = random.choice([x for x in range(nodes) if node_depths[x] < branch])
        length = branch - node_depths[junction]
        edges.append((junction, nodes))
        node_depths.append(node_depths[junction] + 1)
        for nid, new_node in enumerate(range(nodes, nodes + length - 1)):
            edges.append((new_node, new_node + 1))
            node_depths.append(node_depths[junction] + 2 + nid)
        nodes += length
        leaves.append(nodes - 1)

    return edges, node_depths, leaves


def make_objectives(objective_depths, node_depths, leaves):
    """Make num_objectives objectives at leaves with the specified depths.

    ([int], [int]) -> [int]
    """
    objectives = []
    for obj_d in objective_depths:
        candidates = [
            n for n in leaves if node_depths[n] == obj_d and n not in objectives
        ]
        if not candidates:
            raise Exception(
                "Not enough leaf nodes with depth " + str(obj_d) + " for objective."
            )
        objectives.append(random.choice(candidates))

    return objectives


def make_tanks_and_divers(
    node_depths, objectives, num_tank_adjustment, num_diver_adjustment, ordered_tanks
):
    """Make the required number of tank and diver objects.
    |Tanks| = 2^(depth+1). (1 dummy (if ordered))
    |Divers| = 2^(depth-1)

    Adjust the number of tanks and divers by num_tank_adjustment and
    num_diver_adjustment. If either of these numbers is negative, the
    problem will be unsatisfiable.

    ([int], [int], int, int, bool) -> [str], [str]
    """
    num_tanks = num_tank_adjustment - 1
    num_divers = num_diver_adjustment
    for obj in objectives:
        num_tanks += pow(2, node_depths[obj] + 1)
        num_divers += pow(2, node_depths[obj] - 1)
    tanks = ["t" + str(x) for x in range(num_tanks)]
    if ordered_tanks:
        tanks.append("dummy")
    divers = ["d" + str(x) for x in range(num_divers)]

    return tanks, divers


def make_positive_relationships(objectives, node_depths):
    """Create a (transitively closed) graph of relationships showing which
    divers depend on other divers to deliver them tanks to satisfy the
    objectives.

    Process - Start at each objective and walk back to the entrance keeping
    a list of divers as we go. At each step we need 1 additional diver to
    service each diver in our list.

    ([int], [int]) -> set([(int, int)])
    """
    cur_d = 0
    pos_relationships = set()
    for obj in objectives:
        obj_divers = [cur_d]
        cur_d += 1
        depth = node_depths[obj] - 1
        while depth > 0:
            new_divers = list(range(cur_d, cur_d + len(obj_divers)))
            for diver in obj_divers:
                for n_diver in new_divers:
                    pos_relationships.add((n_diver, diver))
            obj_divers.extend(new_divers)
            cur_d += len(new_divers)
            depth -= 1

    return pos_relationships


def make_negative_relationships(pos_relationships, num_divers, neg_link_prob):
    """Make a set of negative relationships where divers preclude each other.

    For the problem to be satisfiable there must just be an ordering over
    the divers that works. Lets assume that the positive relationships
    represent this order. We are then able to rule out everything else.

    In fact, we we have a neg_link_prob chance of ruling out a non-positive
    link.

    (set([(int, int)]), int, ...) -> { int : [int] }
    """
    neg_relationships = {x: list() for x in range(num_divers)}
    for (diver1, diver2) in itertools.combinations(range(num_divers), 2):
        if (
            diver1,
            diver2,
        ) not in pos_relationships and random.random() < neg_link_prob:
            neg_relationships[diver1].append(diver2)

    return neg_relationships


def add_neg_cycle(neg_relationships, num_divers, neg_cycle_frac):
    """Adds a negative cycle to the diver relationships, making the problem
    have no solutions. num_divers * neg_cycle_frac (min 2) divers are
    involved.

    ({ int  : [int] }, int, float) -> None
    """
    divers = random.sample(range(num_divers), max(2, int(num_divers * neg_cycle_frac)))
    for did, diver1 in enumerate(divers):
        diver2 = divers[(did + 1) % len(divers)]
        if diver2 not in neg_relationships[diver1]:
            neg_relationships[diver1].append(diver2)


def make_hiring_costs(neg_relationships, min_cost, max_cost, perturb):
    """Make the hiring costs for the divers. The costs are inversely
    proportional to the number of negative relation ships a diver has.
    They are perturbed by perturb.

    ({ int  : [int] }, int, int, float) -> { int : int }
    """
    divers = list(neg_relationships.keys())
    num_rels = [len(x) for x in list(neg_relationships.values())]
    min_rels, max_rels = min(num_rels), max(num_rels)
    rel_range = max_rels - min_rels
    cost_range = max_cost - min_cost

    if cost_range == 0:
        return {d: min_cost for d in divers}

    if rel_range == 0:
        mid_cost = int(min_cost + cost_range / 2.0)
        return {d: mid_cost for d in divers}

    rel_dict = {n: list() for n in num_rels}
    for nid, num in enumerate(num_rels):
        rel_dict[num].append(divers[nid])
    sorted_rels = sorted(zip(list(rel_dict.keys()), list(rel_dict.values())))
    sorted_rels.reverse()

    hiring_costs = {}
    cost_inc = cost_range / float(len(sorted_rels))
    for rid, (nr, nr_d) in enumerate(sorted_rels):
        for d in nr_d:
            base_cost = min_cost + cost_inc * rid
            base_cost += random.random() * 2 * perturb * base_cost - perturb * base_cost
            base_cost = max(min_cost, min(max_cost, int(base_cost)))
            hiring_costs[d] = base_cost

    return hiring_costs


def write_domain_file(file_name, divers, neg_relationships, strips, ordered_tanks):
    """Write the PDDL domain file to file_name.

    (str, [int], { int : [int] }, bool, bool) -> None
    """
    try:
        output_file = open(file_name, "w")

        if strips:
            output_file.write(";; Cave Diving STRIPS\n")
        else:
            output_file.write(";; Cave Diving ADL\n")
        output_file.write(";; Authors: Nathan Robinson,\n")
        output_file.write(";;          Christian Muise, and\n")
        output_file.write(";;          Charles Gretton\n\n")

        if strips:
            output_file.write("(define (domain cave-diving-strips)\n")
        else:
            output_file.write("(define (domain cave-diving-adl)\n")
        output_file.write("  (:requirements :typing :action-costs :adl)\n")
        output_file.write("  (:types location diver tank quantity)\n")
        output_file.write("  (:predicates\n")
        output_file.write("    (at-tank ?t - tank ?l - location)\n")
        output_file.write("    (in-storage ?t - tank)\n")
        output_file.write("    (full ?t - tank)\n")
        if ordered_tanks:
            output_file.write("    (next-tank ?t1 - tank ?t2 - tank)\n")
        output_file.write("    (at-diver ?d - diver ?l - location)\n")
        output_file.write("    (available ?d - diver)\n")
        output_file.write("    (at-surface ?d - diver)\n")
        output_file.write("    (decompressing ?d - diver)\n")
        if not strips:
            output_file.write("    (precludes ?d1 - diver ?d2 - diver)\n")
        output_file.write("    (cave-entrance ?l - location)\n")
        output_file.write("    (connected ?l1 - location ?l2 - location)\n")
        output_file.write("    (next-quantity ?q1 - quantity ?q2 - quantity)\n")
        output_file.write("    (holding ?d - diver ?t - tank)\n")
        output_file.write("    (capacity ?d - diver ?q - quantity)\n")
        output_file.write("    (have-photo ?l - location)\n")
        output_file.write("    (in-water )\n")

        output_file.write("  )\n\n")

        output_file.write("  (:functions\n")
        output_file.write("    (hiring-cost ?d - diver) - number\n")
        output_file.write("    (other-cost) - number\n")
        output_file.write("    (total-cost) - number\n")
        output_file.write("  )\n\n")

        if not strips:
            output_file.write("  (:action hire-diver\n")
            output_file.write("    :parameters (?d1 - diver)\n")
            output_file.write("    :precondition (and      (available ?d1)\n")
            output_file.write("                       (not (in-water)) \n")
            output_file.write("                  )\n")
            output_file.write("    :effect (and (at-surface ?d1)\n")
            output_file.write("                 (not (available ?d1))\n")
            output_file.write("                 (forall (?d2 - diver)\n")
            output_file.write(
                "                     (when (precludes ?d1 ?d2) (not (available ?d2))))\n"
            )
            output_file.write("                 (in-water)\n")
            output_file.write(
                "                 (increase (total-cost) (hiring-cost ?d1))\n"
            )
            output_file.write("            )\n")
            output_file.write("  )\n\n")
        else:
            for did, diver1 in enumerate(divers):
                output_file.write("  (:action hire-diver-" + diver1 + "\n")
                output_file.write("    :parameters ( )\n")
                output_file.write(
                    "    :precondition (and (available " + diver1 + "))\n"
                )
                output_file.write("    :effect (and (at-surface " + diver1 + ")\n")
                output_file.write("                 (not (available " + diver1 + "))\n")
                for diver2 in neg_relationships[did]:
                    output_file.write(
                        "                 (not (available " + divers[diver2] + "))\n"
                    )
                output_file.write(
                    "                 (increase (total-cost) (hiring-cost "
                    + diver1
                    + "))\n"
                )
                output_file.write("            )\n")
                output_file.write("  )\n\n")

        output_file.write("  (:action prepare-tank\n")
        if ordered_tanks:
            output_file.write(
                "    :parameters (?d - diver ?t1 ?t2 - tank ?q1 ?q2 - quantity)\n"
            )
        else:
            output_file.write(
                "    :parameters (?d - diver ?t1 - tank ?q1 ?q2 - quantity)\n"
            )
        output_file.write("    :precondition (and (at-surface ?d)\n")
        output_file.write("                       (in-storage ?t1)\n")
        output_file.write("                       (next-quantity ?q1 ?q2)\n")
        output_file.write("                       (capacity ?d ?q2)\n")
        if ordered_tanks:
            output_file.write("                       (next-tank ?t1 ?t2)\n")
        output_file.write("                  )\n")
        output_file.write("    :effect (and (not (in-storage ?t1))\n")
        output_file.write("                 (not (capacity ?d ?q2))\n")
        if ordered_tanks:
            output_file.write("                      (in-storage ?t2)\n")
        output_file.write("                      (full ?t1)\n")
        output_file.write("                      (capacity ?d ?q1)\n")
        output_file.write("                      (holding ?d ?t1)\n")
        output_file.write("                 (increase (total-cost) (other-cost ))\n")
        output_file.write("            )\n")
        output_file.write("  )\n\n")

        output_file.write("  (:action enter-water\n")
        output_file.write("    :parameters (?d - diver ?l - location)\n")
        output_file.write("    :precondition (and (at-surface ?d)\n")
        output_file.write("                       (cave-entrance ?l)\n")
        output_file.write("                  )\n")
        output_file.write("    :effect (and (not (at-surface ?d))\n")
        output_file.write("                      (at-diver ?d ?l)\n")
        output_file.write("                 (increase (total-cost) (other-cost ))\n")
        output_file.write("            )\n")
        output_file.write("  )\n\n")

        output_file.write("  (:action pickup-tank\n")
        output_file.write(
            "    :parameters (?d - diver ?t - tank ?l - location ?q1 ?q2 - quantity)\n"
        )
        output_file.write("    :precondition (and (at-diver ?d ?l)\n")
        output_file.write("                       (at-tank ?t ?l)\n")
        output_file.write("                       (next-quantity ?q1 ?q2)\n")
        output_file.write("                       (capacity ?d ?q2)\n")
        output_file.write("                  )\n")
        output_file.write("    :effect (and (not (at-tank ?t ?l))\n")
        output_file.write("                 (not (capacity ?d ?q2))\n")
        output_file.write("                      (holding ?d ?t)\n")
        output_file.write("                      (capacity ?d ?q1)\n")
        output_file.write("                 (increase (total-cost) (other-cost ))\n")
        output_file.write("            )\n")
        output_file.write("  )\n\n")

        output_file.write("  (:action drop-tank\n")
        output_file.write(
            "    :parameters (?d - diver ?t - tank ?l - location ?q1 ?q2 - quantity)\n"
        )
        output_file.write("    :precondition (and (at-diver ?d ?l)\n")
        output_file.write("                       (holding ?d ?t)\n")
        output_file.write("                       (next-quantity ?q1 ?q2)\n")
        output_file.write("                       (capacity ?d ?q1)\n")
        output_file.write("                  )\n")
        output_file.write("    :effect (and (not (holding ?d ?t))\n")
        output_file.write("                 (not (capacity ?d ?q1))\n")
        output_file.write("                      (at-tank ?t ?l)\n")
        output_file.write("                      (capacity ?d ?q2)\n")
        output_file.write("                 (increase (total-cost) (other-cost ))\n")
        output_file.write("            )\n")
        output_file.write("  )\n\n")

        output_file.write("  (:action swim\n")
        output_file.write("    :parameters (?d - diver ?t - tank ?l1 ?l2 - location)\n")
        output_file.write("    :precondition (and (at-diver ?d ?l1)\n")
        output_file.write("                       (holding ?d ?t)\n")
        output_file.write("                       (full ?t)\n")
        output_file.write("                       (connected ?l1 ?l2)\n")
        output_file.write("                  )\n")
        output_file.write("    :effect (and (not (at-diver ?d ?l1))\n")
        output_file.write("                 (not (full ?t))\n")
        output_file.write("                      (at-diver ?d ?l2)\n")
        output_file.write("                 (increase (total-cost) (other-cost ))\n")
        output_file.write("            )\n")
        output_file.write("  )\n\n")

        output_file.write("  (:action photograph\n")
        output_file.write("    :parameters (?d - diver ?l - location ?t - tank)\n")
        output_file.write("    :precondition (and (at-diver ?d ?l)\n")
        output_file.write("                       (holding ?d ?t)\n")
        output_file.write("                       (full ?t)\n")
        output_file.write("                  )\n")
        output_file.write("    :effect (and (not (full ?t))\n")
        output_file.write("                      (have-photo ?l)\n")
        output_file.write("                 (increase (total-cost) (other-cost ))\n")
        output_file.write("            )\n")
        output_file.write("  )\n\n")

        output_file.write("  (:action decompress\n")
        output_file.write("    :parameters (?d - diver ?l - location)\n")
        output_file.write("    :precondition (and (at-diver ?d ?l)\n")
        output_file.write("                       (cave-entrance ?l)\n")
        output_file.write("                  )\n")
        output_file.write("    :effect (and (not (at-diver ?d ?l))\n")
        output_file.write("                      (decompressing ?d)\n")
        output_file.write("                 (not (in-water))\n")
        output_file.write("                 (increase (total-cost) (other-cost ))\n")
        output_file.write("            )\n")
        output_file.write("  )\n\n")

        output_file.write(")\n")
        output_file.close()
    except OSError:
        print("Error: could not write to the domain file:", file_name)


def write_problem_file(
    file_name,
    problem_name,
    num_locations,
    tanks,
    divers,
    objectives,
    edges,
    neg_relationships,
    hiring_costs,
    other_action_cost,
    strips,
    ordered_tanks,
):
    """Write the PDDL problem file to file_name.

    (str, str, int, [str], [str], [int], [(int, int)], { int : [int] },
        { int : int }, int, bool, bool) -> None
    """
    try:
        output_file = open(file_name, "w")

        if strips:
            output_file.write(";; Cave Diving STRIPS\n")
        else:
            output_file.write(";; Cave Diving ADL\n")
        output_file.write(";; Authors: Nathan Robinson,\n")
        output_file.write(";;          Christian Muise, and\n")
        output_file.write(";;          Charles Gretton\n\n")

        if strips:
            output_file.write(
                "(define (problem cave-diving-strips-" + problem_name + ")\n"
            )
            output_file.write("  (:domain cave-diving-strips)\n")
        else:
            output_file.write(
                "(define (problem cave-diving-adl-" + problem_name + ")\n"
            )
            output_file.write("  (:domain cave-diving-adl)\n")

        output_file.write("  (:objects\n")
        output_file.write(
            "    "
            + " ".join(["l" + str(x) for x in range(num_locations)])
            + " - location\n"
        )
        num_diver_lines = len(divers) // 20 + 1
        ordered_divers = ["d" + str(x) for x in range(len(divers))]
        for d_line in range(num_diver_lines):
            output_file.write(
                "    "
                + " ".join(ordered_divers[(d_line * 20) : (d_line * 20 + 20)])
                + " - diver\n"
            )
        num_tank_lines = len(tanks) // 20 + 1
        for t_line in range(num_tank_lines):
            output_file.write(
                "    "
                + " ".join(tanks[(t_line * 20) : (t_line * 20 + 20)])
                + " - tank\n"
            )
        output_file.write("    zero one two three four - quantity\n")
        output_file.write("  )\n\n")

        output_file.write("  (:init\n")
        for diver in ordered_divers:
            output_file.write("    (available " + diver + ")\n")
        for diver in ordered_divers:
            output_file.write("    (capacity " + diver + " four)\n")

        if ordered_tanks:
            output_file.write("    (in-storage " + tanks[0] + ")\n")
            for tid, tank in enumerate(tanks[:-1]):
                output_file.write(
                    "    (next-tank " + tank + " " + tanks[tid + 1] + ")\n"
                )
        else:
            for tank in tanks:
                output_file.write("    (in-storage " + tank + ")\n")

        output_file.write("    (cave-entrance l0)\n")

        for edge in edges:
            output_file.write(
                "    (connected l" + str(edge[0]) + " l" + str(edge[1]) + ")\n"
            )
            output_file.write(
                "    (connected l" + str(edge[1]) + " l" + str(edge[0]) + ")\n"
            )

        output_file.write("    (next-quantity zero one)\n")
        output_file.write("    (next-quantity one two)\n")
        output_file.write("    (next-quantity two three)\n")
        output_file.write("    (next-quantity three four)\n")

        if not strips:
            for did1, diver1 in enumerate(divers):
                for diver2 in neg_relationships[did1]:
                    output_file.write(
                        "    (precludes " + diver1 + " " + divers[diver2] + ")\n"
                    )

        for did, diver in enumerate(divers):
            output_file.write(
                "    (= (hiring-cost " + diver + ") " + str(hiring_costs[did]) + ")\n"
            )

        output_file.write("    (= (other-cost ) " + str(other_action_cost) + ")\n")

        output_file.write("    (= (total-cost) 0)\n")
        output_file.write("  )\n\n")

        output_file.write("  (:goal\n")
        output_file.write("    (and\n")
        for obj in objectives:
            output_file.write("      (have-photo l" + str(obj) + ")\n")
        for diver in divers:
            output_file.write("      (decompressing " + diver + ")\n")
        output_file.write("    )\n  )\n\n")
        output_file.write("  (:metric minimize (total-cost))\n\n")
        output_file.write(")\n")

        output_file.close()
    except OSError:
        print("Error: could not write to the problem file:", file_name)


def main():
    args = generator_cmd_line.process_args()
    random.seed(args.seed)

    edges, node_depths, leaves = make_caves(args.cave_branches)

    objectives = make_objectives(args.objectives, node_depths, leaves)

    tanks, divers = make_tanks_and_divers(
        node_depths,
        objectives,
        args.num_tank_adjustment,
        args.num_diver_adjustment,
        args.order_tanks,
    )

    pos_relationships = make_positive_relationships(objectives, node_depths)

    neg_relationships = make_negative_relationships(
        pos_relationships, len(divers), args.neg_link_prob
    )

    if args.neg_cycle_length:
        add_neg_cycle(neg_relationships, len(divers), args.neg_cycle_length)

    hiring_costs = make_hiring_costs(
        neg_relationships,
        args.minimum_hiring_cost,
        args.maximum_hiring_cost,
        args.perturb_hiring_costs,
    )

    random.shuffle(divers)

    if not args.quiet:
        print()
        print("Edges: ", ", ".join(map(str, edges)))
        print("Depths:", ", ".join(map(str, node_depths)))
        print("Objectives:", ", ".join(map(str, objectives)))
        print("Tanks:", ", ".join(map(str, tanks)))
        print("Divers:", ", ".join(map(str, divers)))
        print("Positive relationships:", ", ".join(map(str, pos_relationships)))
        print("Negative relationships:", neg_relationships)
        print("Hiring costs:", hiring_costs)

    if args.domain_file_name:
        write_domain_file(
            args.domain_file_name,
            divers,
            neg_relationships,
            args.strips,
            args.order_tanks,
        )

    write_problem_file(
        args.problem_file_name,
        args.problem_name,
        len(node_depths),
        tanks,
        divers,
        objectives,
        edges,
        neg_relationships,
        hiring_costs,
        args.other_action_cost,
        args.strips,
        args.order_tanks,
    )


if __name__ == "__main__":
    main()



#! /usr/bin/env python

import argparse
from collections import deque
from copy import deepcopy
import logging
from pathlib import Path
import random
import sys
from typing import Optional, Set


class Map:
    def __init__(self, nodes: int):
        self.nodes = nodes
        self.edges = [[] for n in range(self.nodes)]

    def add_edge(self, src: int, dst: int):
        if dst not in self.edges[src]:
            self.edges[src].append(dst)
            self.edges[dst].append(src)

    def remove_edges_at(self, node: int):
        assert 0 <= node and node <= self.nodes
        for i, edges in enumerate(self.edges):
            self.edges[i] = [dst for dst in edges if dst != node]
        self.edges[node] = []

    def clone(self):
        return deepcopy(self)

    # do BFS exploration
    def path(self, init: int, goal: int):
        assert 0 <= init and init < self.nodes and goal < self.nodes
        q = deque()
        q.append(init)
        explored = set([init])
        parent = [None for _ in range(self.nodes)]
        while len(q) > 0:
            src = q.popleft()
            assert src in explored
            if src == goal: break
            for dst in random.sample(self.edges[src], len(self.edges[src])):
                if dst not in explored:
                    explored.add(dst)
                    parent[dst] = src
                    q.append(dst)

        if goal < 0:
            return explored
        elif parent[goal] is None:
            return None
        else:
            src = goal
            rpath = []
            while src != init:
                rpath.append(src)
                src = parent[src]
                assert src is not None
            rpath.append(init)
            return list(reversed(rpath))

    def reachable_nodes(self, init: int):
        return self.path(init, -1)

class Floorplan:
    def __init__(self, fname: Path, logger):
        # read template file
        with fname.open('r') as fd:
            self.lines = [line.strip('\n') for line in fd.readlines()]

        # initialize data structures
        self.cells, self.cmap = [], dict()
        self.locks = set()
        self.logger = logger
        self.robot = None
        self.goal = None

        # first pass, read cells
        for r, line in enumerate(self.lines):
            for c, char in enumerate(line):
                if char not in [' ', 'w']:
                    cell = (c, r)
                    self.cmap[cell] = len(self.cells)
                    self.cells.append(cell)
                    if char == 'L': self.locks.add(cell)
                    if char == 'R': self.robot = cell
                    if char == 'G': self.goal = cell

        # second pass, read connections
        self.map = Map(len(self.cells))
        for (c, r) in self.cells:
            cell = (c, r)
            src = self.cmap[cell]
            leftcell = (c-1, r)
            if leftcell in self.cmap:
                dst = self.cmap[leftcell]
                self.map.add_edge(src, dst)
            abovecell = (c, r-1)
            if abovecell in self.cmap:
                dst = self.cmap[abovecell]
                self.map.add_edge(src, dst)
        logger.info(f'Floorplan: ncells={len(self.cells)}, locks={self.locks}')

    def get_map(self, locks: Optional[Set] = None):
        map = self.map.clone()
        if locks != None:
            for lock in locks:
                map.remove_edges_at(self.cmap[lock])
        return map


class Instance:
    def __init__(self, name: str, floorplan: Floorplan, nshapes: int, logger):
        self.name = name
        self.logger = logger
        self.floorplan = floorplan

        # assign shapes to locks, there would be one key per shape
        nlocks = len(floorplan.locks)
        if nshapes == 0:
            nshapes = random.randint(1, nlocks) if nlocks > 0 else 0
        self.lock_shapes = list(map(lambda x: x, random.choices(range(nshapes), k=nlocks)))
        self.key_shapes = list(range(0, nshapes))

        self.shape_map = dict()
        for i, cell in enumerate(floorplan.locks):
            self.shape_map[cell] = self.lock_shapes[i]
        logger.info(f'Locks: nlocks={nlocks}, nshapes={nshapes}, shape_map={self.shape_map}')

        # place robot and goal
        free_cells = list(set(floorplan.cells) - floorplan.locks)
        self.robot = random.sample(free_cells, k=1)[0] if floorplan.robot is None else floorplan.robot
        if floorplan.goal is None:
            reachable_cells = floorplan.map.reachable_nodes(floorplan.cmap[self.robot])
            self.goal = floorplan.cells[random.sample(list(reachable_cells), k=1)[0]]
        else:
            self.goal = floorplan.goal
        logger.debug(f'Init/goal: robot={self.robot}, goal={self.goal}')

        # construct high-level path from robot to goal
        robot_to_goal_path = floorplan.map.path(floorplan.cmap[self.robot], floorplan.cmap[self.goal])
        assert robot_to_goal_path is not None, f"With all locks open, cell {self.goal} isn't reachable from {self.robot}"
        lock_path = [floorplan.cells[cell] for cell in robot_to_goal_path if floorplan.cells[cell] in floorplan.locks]
        logger.info(f'Path: init={self.robot}, goal={self.goal}, path={list(map(lambda i: floorplan.cells[i], robot_to_goal_path))}, locks={lock_path}')

        # place keys so that resulting placement is solvable
        locks = set(floorplan.locks)
        self.key_locations = [None for _ in range(nshapes)] # one key per shape
        current_map = floorplan.get_map(locks)
        current_pos = floorplan.cmap[self.robot]
        previous_loc = set([current_pos])
        for i, lock in enumerate(lock_path):
            shape = self.shape_map[lock]
            logger.debug(f'i={i}, current_pos={floorplan.cells[current_pos]}, lock={lock}, shape={shape}')
            if self.key_locations[shape] == None:
                reachable_cells = current_map.reachable_nodes(current_pos) - previous_loc
                self.key_locations[shape] = random.sample(list(reachable_cells), k=1)[0]
                logger.debug(f'key_location: shape={shape}, loc={floorplan.cells[self.key_locations[shape]]}, reachable_cells={list(map(lambda i: floorplan.cells[i], reachable_cells))}')
                current_pos = self.key_locations[shape]
                previous_loc.add(current_pos)
                open_locks = set([lock for lock in floorplan.locks if self.shape_map[lock] == shape])
                logger.debug(f'open_locks={open_locks}')
                locks -= open_locks
                current_map = floorplan.get_map(locks)

        # keys that have not yet placed can be anywhere
        for i, loc in enumerate(self.key_locations):
            if loc is None:
                self.key_locations[i] = floorplan.cmap[random.sample(list(floorplan.cells), k=1)[0]]

    def write(self, fname: Path):
        with fname.open('w') as fd:
            fd.write(f'(define (problem {self.name})\n')
            fd.write(f'  (:domain grid)\n')

            # objects
            fd.write(f'  (:objects\n')
            fd.write(f'   ')
            for i, (c, r) in enumerate(self.floorplan.cells):
                fd.write(f' p{i}')
            fd.write(f'\n   ')
            for i, shape in enumerate(self.key_shapes):
                fd.write(f' shape{shape}')
            fd.write(f'\n   ')
            for i, shape in enumerate(self.key_shapes):
                fd.write(f' key{i}')
            fd.write('\n  )\n')

            # init: types
            fd.write(f'  (:init')
            fd.write(f'\n    ; Object types\n   ')
            for i, (c, r) in enumerate(self.floorplan.cells):
                fd.write(f' (place p{i})')
            fd.write(f'\n   ')
            for i, shape in enumerate(self.key_shapes):
                fd.write(f' (shape shape{shape})')
            fd.write(f'\n   ')
            for i, shape in enumerate(self.key_shapes):
                fd.write(f' (key key{i})')

            # init: open/locked cells
            fd.write(f'\n    ; Open/locked cells\n   ')
            for i, (c, r) in enumerate(self.floorplan.cells):
                if (c, r) not in self.floorplan.locks:
                    fd.write(f' (open p{i})')
            if len(self.floorplan.locks) > 0:
                fd.write(f'\n   ')
                for i, (c, r) in enumerate(self.floorplan.cells):
                    if (c, r) in self.floorplan.locks:
                        fd.write(f' (locked p{i})')
            fd.write('\n')

            # init: connected cells
            fd.write(f'    ; Connected cells\n')
            for src in range(self.floorplan.map.nodes):
                for dst in self.floorplan.map.edges[src]:
                    fd.write(f'    (conn p{src} p{dst})\n')

            # init: assign shapes
            fd.write(f'    ; Lock and key shapes\n')
            for cell in self.shape_map:
                fd.write(f'    (lock-shape p{self.floorplan.cmap[cell]} shape{self.shape_map[cell]})\n')
            for i, shape in enumerate(self.key_shapes):
                fd.write(f'    (key-shape key{i} shape{shape})\n')

            # init: place keys
            fd.write(f'    ; Key placement\n')
            for i, loc in enumerate(self.key_locations):
                fd.write(f'    (at key{i} p{loc})\n')

            # init: robot
            fd.write(f'    ; Robot placement\n')
            fd.write(f'    (at-robot p{self.floorplan.cmap[self.robot]})\n')
            fd.write(f'    (arm-empty)\n')
            fd.write(f'  )\n')

            # goal
            fd.write(f'  (:goal (at-robot p{self.floorplan.cmap[self.goal]}))\n')
            fd.write(f')\n')


def get_logger(name: str, log_file: Path, level = logging.INFO):
    logger = logging.getLogger(name)
    logger.propagate = False
    logger.setLevel(level)

    # add stdout handler
    #formatter = logging.Formatter('[%(levelname)s] %(message)s')
    formatter = logging.Formatter('%(asctime)s [%(levelname)s] [%(funcName)s:%(lineno)d] %(message)s')
    console = logging.StreamHandler(sys.stdout)
    console.setFormatter(formatter)
    logger.addHandler(console)

    # add file handler
    if log_file != '':
        formatter = logging.Formatter('%(asctime)s [%(levelname)s] [%(funcName)s:%(lineno)d] %(message)s')
        file_handler = logging.FileHandler(str(log_file), 'a')
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)

    return logger

def close_logger(logger):
    handlers = logger.handlers
    for handler in handlers:
       logger.removeHandler(handler)
       handler.close()

def get_args():
    # argument parser
    default_seed = 0
    default_debug_level = 0
    default_num_instances = 1
    parser = argparse.ArgumentParser('mini_grid.py')
    parser.add_argument('--seed', type=int, default=default_seed, help=f'Seed for random generator (default={default_seed})')
    parser.add_argument('--debug_level', type=int, default=default_debug_level, help=f'Set debug level (default={default_debug_level})')
    parser.add_argument('--num_instances', type=int, default=default_num_instances, help=f'Number of instances to generate (default={default_num_instances})')
    parser.add_argument('floorplan', type=Path, help='Filename for floorplan')
    parser.add_argument('nshapes', type=int, help='Number of shapes for locks (0 means choose it randomly)')

    # paths
    default_results_path = ''
    default_floorplans_path = ''
    paths = parser.add_argument_group('paths')
    paths.add_argument('--results', type=Path, default=default_results_path, help=f"Path to results folders (default='{default_results_path}')")
    paths.add_argument('--floorplans_path', type=Path, default=default_floorplans_path, help=f"Path to floorplans (default='{default_floorplans_path}')")

    # parse arguments
    args = parser.parse_args()
    return args

def main(args: argparse.Namespace):
    # setup logger and identify call
    log_file = Path('./log.txt')
    log_level = logging.INFO if args.debug_level == 0 else logging.DEBUG
    logger = get_logger('mini_grid.py', log_file, log_level)
    logger.info(f"Using log file '{log_file}'")
    logger.info(f'call=|{" ".join(sys.argv)}|')

    # set random seed
    random.seed(args.seed)

    # read floorplan
    floorplan = Floorplan(args.floorplans_path / args.floorplan, logger)

    # generate random instances
    name = args.floorplan.name
    args.results.mkdir(parents=True, exist_ok=True)
    for i in range(args.num_instances):
        name = f'grid_{name}_s{args.nshapes}_seed{args.seed}_n{i}'.replace('.', '_')
        pddl_name = args.results / Path(name).with_suffix('.pddl')
        instance = Instance(name, floorplan, args.nshapes, logger)
        instance.write(pddl_name)
        logger.info(f'{pddl_name} written!')


if __name__ == '__main__':
    # read arguments and do job
    args = get_args()
    main(args)




#!/usr/bin/env python

import random
import sys

MAX_LEVELS = 2


def print_usage():
    print(
        "Usage: "
        + sys.argv[0]
        + " <num_cocktails> <num_ingredients> <num_shots> [<random_seed>]"
    )
    print("  num_cocktails (min 1)")
    print("  num_ingredients (min 2)")
    print("  num_shots (min num_cocktails+1)")
    print("  random_seed (min 1, optional)")


if len(sys.argv) not in [4, 5]:
    print_usage()
    sys.exit(2)

num_cocktails = int(sys.argv[1])
num_ingredients = int(sys.argv[2])
num_shots = int(sys.argv[3])

if len(sys.argv) == 5:
    seed = int(sys.argv[4])
else:
    seed = None


def get_objects():
    str_objects = "\n"

    # -- shaker
    str_objects = str_objects + "      shaker1 - shaker\n"

    # -- robot hands
    str_objects = str_objects + "      left right - hand\n     "

    # -- shots
    for i in range(1, num_shots + 1):
        str_objects = str_objects + " shot" + str(i)
    str_objects = str_objects + " - shot\n     "

    # -- ingredients
    for i in range(1, num_ingredients + 1):
        str_objects = str_objects + " ingredient" + str(i)
    str_objects = str_objects + " - ingredient\n     "

    # -- mixs
    for i in range(1, num_cocktails + 1):
        str_objects = str_objects + " cocktail" + str(i)
    str_objects = str_objects + " - cocktail\n     "

    # -- dispensers
    for i in range(1, num_ingredients + 1):
        str_objects = str_objects + " dispenser" + str(i)
    str_objects = str_objects + " - dispenser\n     "

    # -- levels
    for i in range(MAX_LEVELS + 1):
        str_objects = str_objects + " l" + str(i)
    str_objects = str_objects + " - level\n"

    return str_objects


def get_init():
    str_init = "\n"
    str_init = str_init + "  (ontable shaker1)\n"

    for i in range(1, num_shots + 1):
        str_init = str_init + "  (ontable shot" + str(i) + ")\n"

    for i in range(1, num_ingredients + 1):
        str_init = (
            str_init
            + "  (dispenses dispenser"
            + str(i)
            + " ingredient"
            + str(i)
            + ")\n"
        )

    str_init = str_init + "  (clean shaker1)\n"

    for i in range(1, num_shots + 1):
        str_init = str_init + "  (clean shot" + str(i) + ")\n"

    str_init = str_init + "  (empty shaker1)\n"

    for i in range(1, num_shots + 1):
        str_init = str_init + "  (empty shot" + str(i) + ")\n"

    str_init = str_init + "  (handempty left)\n  (handempty right)\n"

    str_init = str_init + "  (shaker-empty-level shaker1 l0)\n"
    str_init = str_init + "  (shaker-level shaker1 l0)\n"
    for i in range(MAX_LEVELS):
        str_init = str_init + "  (next l" + str(i) + " l" + str(i + 1) + ")\n"

    for i in range(1, num_cocktails + 1):
        parts = random.sample(list(range(1, num_ingredients + 1)), 2)
        str_init = (
            str_init
            + "  (cocktail-part1 cocktail"
            + str(i)
            + " ingredient"
            + str(parts[0])
            + ")\n"
        )
        str_init = (
            str_init
            + "  (cocktail-part2 cocktail"
            + str(i)
            + " ingredient"
            + str(parts[1])
            + ")\n"
        )
    return str_init


def get_goals():
    str_goal = ""
    str_goal = str_goal + "\n  (and\n"

    serving = random.sample(list(range(1, num_cocktails + 1)), num_cocktails)
    for i in range(1, num_cocktails + 1):
        str_goal = (
            str_goal
            + "      (contains shot"
            + str(i)
            + " cocktail"
            + str(serving[i - 1])
            + ")\n"
        )

    # there is at least one shot for not serving
    for j in range(i + 1, num_shots):
        flip = random.randint(0, 1)
        if flip == 1:
            str_goal = (
                str_goal
                + "      (contains shot"
                + str(j)
                + " cocktail"
                + str(random.randint(1, num_cocktails))
                + ")\n"
            )
        else:
            str_goal = (
                str_goal
                + "      (contains shot"
                + str(j)
                + " ingredient"
                + str(random.randint(1, num_ingredients))
                + ")\n"
            )

    str_goal = str_goal + ")"
    return str_goal


def main():
    name = "prob"

    random.seed(seed)

    print("(define (problem " + name + ")")
    print(" (:domain barman)")
    print(" (:objects " + get_objects() + ")")
    print(" (:init " + get_init() + ")")
    print(" (:goal" + get_goals() + "))")


main()



import sys
sys.path.append("/Users/omarwattad/Documents/Action Model - Research/sam_learning")
from sam_learning.learners.numeric_sam import NumericSAMLearner
from sam_learning.learners.numeric_sam import *
from sam_learning.core import NumericFluentStateStorage
from typing import List, Dict, Tuple, Optional
sys.path.append("/Users/omarwattad/Documents/Action Model - Research/sam_learning/utilities")
from utilities import NegativePreconditionPolicy
from pddl_plus_parser.models import Observation, ActionCall, State, Domain, Precondition


class OnlyNumericSam(NumericSAMLearner):
    def __init__(
        self,
        partial_domain: Domain,
        relevant_fluents: Optional[Dict[str, List[str]]] = None,
        allow_unsafe: bool = False,
        polynomial_degree: int = 0,
        negative_preconditions_policy: NegativePreconditionPolicy = NegativePreconditionPolicy.no_remove,
        clean_data: bool = True,
        **kwargs,
    ):
        super().__init__(partial_domain, relevant_fluents, allow_unsafe, polynomial_degree, negative_preconditions_policy=negative_preconditions_policy)
        if not clean_data:
            self.learned_data = partial_domain
        else: self.learned_data = None

    def add_data_to_domain(self): # todo: CHECK IF NEEDED!
        if self.learned_data:
            for action_str , action in self.learned_data.actions.items():
                for condition in action.preconditions.root.operands:
                    self.partial_domain.actions[action_str].preconditions.add_condition(condition)
                self.partial_domain.actions[action_str].discrete_effects.update(action.discrete_effects)

    def add_new_action(self, grounded_action: ActionCall, previous_state: State, next_state: State) -> None:
        """Adds a new action to the learned domain.

        :param grounded_action: the grounded action that was executed according to the observation.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous
        """
        # super().add_new_action(grounded_action, previous_state, next_state) # SAM Learner
        self.logger.debug(f"Creating the new storage for the action - {grounded_action.name}.")
        previous_state_lifted_matches = self.function_matcher.match_state_functions(grounded_action, self.triplet_snapshot.previous_state_functions)
        next_state_lifted_matches = self.function_matcher.match_state_functions(grounded_action, self.triplet_snapshot.next_state_functions)
        possible_bounded_functions = self.vocabulary_creator.create_lifted_functions_vocabulary(
            domain=self.partial_domain, possible_parameters=self.partial_domain.actions[grounded_action.name].signature
        )
        self.storage[grounded_action.name] = NumericFluentStateStorage(
            action_name=grounded_action.name, domain_functions=possible_bounded_functions, polynom_degree=self.polynom_degree,
        )
        observed_action = self.partial_domain.actions[grounded_action.name]
        self.observed_actions.append(observed_action.name)
        self.storage[grounded_action.name].add_to_previous_state_storage(previous_state_lifted_matches)
        self.storage[grounded_action.name].add_to_next_state_storage(next_state_lifted_matches)
        self.logger.debug(f"Done creating the numeric state variable storage for the action - {grounded_action.name}")

    def update_action(self, grounded_action: ActionCall, previous_state: State, next_state: State) -> None:
        """Updates the action's data according to the new input observed triplet.

        :param grounded_action: the grounded action that was observed.
        :param previous_state: the state that the action was executed on.
        :param next_state: the state that was created after executing the action on the previous
            state.
        """
        action_name = grounded_action.name
        # super().update_action(grounded_action, previous_state, next_state) # SAM learner!
        self.logger.debug(f"Adding the numeric state variables to the numeric storage of action - {action_name}.")
        previous_state_lifted_matches = self.function_matcher.match_state_functions(grounded_action, self.triplet_snapshot.previous_state_functions)
        next_state_lifted_matches = self.function_matcher.match_state_functions(grounded_action, self.triplet_snapshot.next_state_functions)
        self.storage[action_name].add_to_previous_state_storage(previous_state_lifted_matches)
        self.storage[action_name].add_to_next_state_storage(next_state_lifted_matches)
        self.logger.debug(f"Done updating the numeric state variable storage for the action - {grounded_action.name}")

    def learn_action_model(self, observations: List[Observation]) -> Tuple[LearnerDomain, Dict[str, str]]:
        """Learn the SAFE action model from the input observations.

        :param observations: the list of trajectories that are used to learn the safe action model.
        :return: a domain containing the actions that were learned and the metadata about the learning.
        """
        self.logger.info("Starting to learn the action model!")
        super().start_measure_learning_time()
        super().deduce_initial_inequality_preconditions()
        self.add_data_to_domain()
        for observation in observations:
            self.current_trajectory_objects = observation.grounded_objects
            for component in observation.components:
                if not super().are_states_different(component.previous_state, component.next_state):
                    continue

                self.handle_single_trajectory_component(component)

        self.handle_negative_preconditions_policy()
        allowed_actions, learning_metadata = self._create_safe_action_model()

        super().end_measure_learning_time()
        learning_metadata["learning_time"] = str(self.learning_end_time - self.learning_start_time)
        return self.partial_domain, learning_metadata


r"""
Syntactic similarity metrics compare the intersection or difference of the atoms
in the action preconditions and effects between an evaluated domain model :math:`M`
and reference model :math:`M^{*}`.
Let :math:`pre_M(a)` be the set of preconditions of an action :math:`a` according to :math:`M`.

* True Positives: :math:`TP_{pre}(a)=|(pre_M(a)\cap pre_M^{*}(a)|`
* False Positives: :math:`FP_{pre}(a)=|(pre_M(a)\setminus pre_M^{*}(a)|`
* False Negatives: :math:`FN_{pre}(a)=|(pre_M^{*}(a)\setminus pre_M(a))|`

The preconditions precision :math:`P_{pre}` and recall :math:`R_{pre}` are defined as:

* Syntactic precision of :math:`pre_a` : :math:`P_{pre}(a) = \frac{TP_{pre}(a)}{TP_{pre}(a)+FP_{pre}(a)}`
* Syntactic recall of :math:`pre_a`: :math:`R_{pre}(a) = \frac{TP_{pre}(a)}{TP_{pre}(a)+FN_{pre}(a)}`

The precision of an action :math:`a` is defined by summing up :math:`TP(a)` over
the preconditions :math:`pre`, positive effects :math:`eff^+` and negative effects :math:`eff^-`:

* Syntactic precision of :math:`a` : :math:`P(a) = \frac{TP(a)}{TP(a)+FP(a)}`
* Syntactic recall of :math:`a` : :math:`R(a) = \frac{TP(a)}{TP(a)+FN(a)}`

Finally, the overall precision and recall of a domain model averages over the actions precision and recall:

* **Syntactic precision** of :math:`M`: :math:`P = \frac{1}{|A|}\sum_{a\in A} P(a)`
* **Syntactic recall** of :math:`M`: :math:`R = \frac{1}{|A|}\sum_{a\in A} R(a)`


"""

from amlgym.util.SimpleDomainReader import SimpleDomainReader, Operator

from typing import Dict
import numpy as np
import warnings
import copy
from sklearn.metrics import precision_score, recall_score


def syntactic_precision(evaluated_path: str,
                        reference_path: str) -> Dict[str, float]:
    """
    Evaluate the syntactic precision metric of a domain model :math:`M` with
    respect to a reference model :math:`M^{*}`.

    :param evaluated_path: path of the PDDL model to evaluate.
    :param reference_path: path of the PDDL reference model.

    :return: the syntactic precision grouped by preconditions and effects
    """

    eval_model = SimpleDomainReader(input_file=evaluated_path)
    ref_model = SimpleDomainReader(input_file=reference_path)

    # Normalize operator names
    for op_gt, op_eval in zip(ref_model.operators, eval_model.operators):
        op_gt.operator_name = op_gt.operator_name.replace('_', '-')
        op_eval.operator_name = op_eval.operator_name.replace('_', '-')

    # Sort reference model operators
    ref_model.operators.sort(key=lambda op: op.operator_name, reverse=True)
    eval_operator_map = {op.operator_name: op for op in eval_model.operators}

    # Align evaluated operators to the reference order
    eval_model.operators = [
        eval_operator_map.get(op.operator_name, _empty_operator_like(op))
        for op in ref_model.operators
    ]

    # Verify correct operator alignment
    assert all(
        gt_op.operator_name == eval_op.operator_name
        for gt_op, eval_op in zip(ref_model.operators, eval_model.operators)
    )

    # Measure the preconditions/effects precision for every operator
    results = {k: [] for k in ["precs_pos", "precs_neg", "eff_pos", "eff_neg", "overall"]}

    for gt_op, eval_op in zip(ref_model.operators, eval_model.operators):
        metrics = _compute_operator_precision(gt_op, eval_op)
        for key, val in metrics.items():
            results[key].append(val)

    # Aggregate results
    return {
        'precs_pos': np.round(np.mean(results["precs_pos"]), 2),
        'precs_neg': np.round(np.mean(results["precs_neg"]), 2),
        'eff_pos': np.round(np.mean(results["eff_pos"]), 2),
        'eff_neg': np.round(np.mean(results["eff_neg"]), 2),
        'mean': np.round(np.mean(results["overall"]), 2)
    }


def syntactic_recall(evaluated_path: str,
                     reference_path: str) -> Dict[str, float]:
    """
    Evaluate the syntactic recall metric of a domain model :math:`M` with
    respect to a reference model :math:`M^{*}`.

    :param evaluated_path: path of the PDDL model to evaluate.
    :param reference_path: path of the PDDL reference model.

    :return: the syntactic recall grouped by preconditions and effects
    """

    eval_model = SimpleDomainReader(input_file=evaluated_path)
    ref_model = SimpleDomainReader(input_file=reference_path)

    # Normalize operator names
    for op_gt, op_eval in zip(ref_model.operators, eval_model.operators):
        op_gt.operator_name = op_gt.operator_name.replace('_', '-')
        op_eval.operator_name = op_eval.operator_name.replace('_', '-')

    # Sort reference model operators
    ref_model.operators.sort(key=lambda op: op.operator_name, reverse=True)
    eval_operator_map = {op.operator_name: op for op in eval_model.operators}

    # Align evaluated operators to the reference order
    eval_model.operators = [
        eval_operator_map.get(op.operator_name, _empty_operator_like(op))
        for op in ref_model.operators
    ]

    # Verify correct operator alignment
    assert all(
        gt_op.operator_name == eval_op.operator_name
        for gt_op, eval_op in zip(ref_model.operators, eval_model.operators)
    )

    # Measure the preconditions/effects precision for every operator
    results = {k: [] for k in ["precs_pos", "precs_neg", "eff_pos", "eff_neg", "overall"]}

    for gt_op, eval_op in zip(ref_model.operators, eval_model.operators):
        metrics = _compute_operator_recall(gt_op, eval_op)
        for key, val in metrics.items():
            results[key].append(val)

    # Aggregate results
    return {
        'precs_pos': np.round(np.mean(results["precs_pos"]), 2),
        'precs_neg': np.round(np.mean(results["precs_neg"]), 2),
        'eff_pos': np.round(np.mean(results["eff_pos"]), 2),
        'eff_neg': np.round(np.mean(results["eff_neg"]), 2),
        'mean': np.round(np.mean(results["overall"]), 2)
    }


def _empty_operator_like(op: Operator) -> Operator:
    """
    Return a copy of op with all preconditions and effects empty.

    :param op: input operator

    :return: a copy of `op` with no preconditions and no effects
    """
    new_op = copy.deepcopy(op)
    new_op.precs_pos = []
    new_op.precs_neg = []
    new_op.eff_pos = []
    new_op.eff_neg = []
    return new_op


def _compute_operator_precision(reference_op: Operator,
                                evaluated_op: Operator) -> Dict[str, float]:
    """
    Compute syntactic precision of preconditions and effects of a single operator.

    :param reference_op: ground-truth operator
    :param evaluated_op: operator to be evaluated

    :return: operator syntactic precision grouped by preconditions and effects
    """
    categories = {
        "precs_pos": (reference_op.precs_pos, evaluated_op.precs_pos),
        "precs_neg": (reference_op.precs_neg, evaluated_op.precs_neg),
        "eff_pos": (reference_op.eff_pos, evaluated_op.eff_pos),
        "eff_neg": (reference_op.eff_neg, evaluated_op.eff_neg),
    }

    precisions = {}
    all_tp = all_fp = all_fn = 0

    for key, (gt, pred) in categories.items():
        universe = list(set(gt) | set(pred))
        y_true = [1 if p in gt else 0 for p in universe]
        y_pred = [1 if p in pred else 0 for p in universe]

        if not universe:
            # No predicates at all
            warnings.warn(f"No {key} for operator {reference_op.operator_name}, "
                          f"precision set to 1.", stacklevel=2)
            precision = 1.0
        else:
            precision = precision_score(y_true, y_pred, zero_division=1.)

        precisions[key] = precision

        # Compute tp/fp/fn counts for overall precision
        gt_set, pred_set = set(gt), set(pred)
        all_tp += len(gt_set & pred_set)
        all_fp += len(pred_set - gt_set)
        all_fn += len(gt_set - pred_set)

    overall_precision = (
        all_tp / (all_tp + all_fp)
        if (all_tp + all_fp) > 0
        else 1.0
    )

    precisions["overall"] = overall_precision
    return precisions


def _compute_operator_recall(reference_op: Operator,
                             evaluated_op: Operator) -> Dict[str, float]:
    """
    Compute syntactic recall of preconditions and effects of a single operator.

    :param reference_op: ground-truth operator
    :param evaluated_op: operator to be evaluated

    :return: operator syntactic recall grouped by preconditions and effects
    """
    categories = {
        "precs_pos": (reference_op.precs_pos, evaluated_op.precs_pos),
        "precs_neg": (reference_op.precs_neg, evaluated_op.precs_neg),
        "eff_pos": (reference_op.eff_pos, evaluated_op.eff_pos),
        "eff_neg": (reference_op.eff_neg, evaluated_op.eff_neg),
    }

    recalls = {}
    all_tp = all_fp = all_fn = 0

    for key, (gt, pred) in categories.items():
        universe = list(set(gt) | set(pred))
        y_true = [1 if p in gt else 0 for p in universe]
        y_pred = [1 if p in pred else 0 for p in universe]

        if not universe:
            # No predicates at all
            warnings.warn(f"No {key} for operator {reference_op.operator_name}, "
                          f"recall set to 1.", stacklevel=2)
            recall = 1.0
        else:
            recall = recall_score(y_true, y_pred, zero_division=1.)

        recalls[key] = recall

        # Compute tp/fp/fn counts for overall recall
        gt_set, pred_set = set(gt), set(pred)
        all_tp += len(gt_set & pred_set)
        all_fp += len(pred_set - gt_set)
        all_fn += len(gt_set - pred_set)

    overall_recall = (
        all_tp / (all_tp + all_fn)
        if (all_tp + all_fn) > 0
        else 1.0
    )

    recalls["overall"] = overall_recall
    return recalls



import re
import os

from amlgym.algorithms.OfflineAlgorithmAdapter import OfflineAlgorithmAdapter
from typing import List, ClassVar, OrderedDict
import shutil
from pathlib import Path
from pddl_plus_parser.lisp_parsers import DomainParser, TrajectoryParser
from pddl_plus_parser.lisp_parsers import ProblemParser

from amlgym.algorithms.rosame.experiment_runner.rosame_runner import Rosame_Runner


class ROSAME(OfflineAlgorithmAdapter):
    """
    Adapter class for running an *unofficial* implementation of the ROSAME
    algorithm: "Neuro-Symbolic Learning of Lifted Action Models from
    Visual Traces", Kai Xi, Stephen Gould, Sylvie Thiebaux, Proceedings of the
    Thirty-Fourth International Conference on Automated Planning and Scheduling, 2024.
    https://ojs.aaai.org/index.php/ICAPS/article/download/31528/33688

    Example:
        .. code-block:: python

            from amlgym.algorithms import get_algorithm
            rosame = get_algorithm('ROSAME')
            model = rosame.learn('path/to/domain.pddl', ['path/to/trace0', 'path/to/trace1'])
            print(model)

    """
    _reference: ClassVar[OrderedDict[str, str]] = {
        'Authors': "K. Xi, S. Gould, and S. Thiebaux",
        'Title': "Neuro-Symbolic Learning of Lifted Action Models from Visual Traces",
        'Venue': "International Conference on Automated Planning and Scheduling",
        'Year': 2024,
        'URL': "https://ojs.aaai.org/index.php/ICAPS/article/download/31528/33688"
    }

    def __init__(self, **kwargs):
        super(ROSAME, self).__init__(**kwargs)

    def learn(self,
              domain_path: str,
              trajectory_paths: List[str],
              use_problems: bool = True) -> str:
        """
        Learns a PDDL action model from:
         (i)    a (possibly empty) input model which is required to specify the predicates and operators signature;
         (ii)   a list of trajectory file paths.

        :parameter domain_path: input PDDL domain file path
        :parameter trajectory_paths: list of trajectory file paths
        :parameter use_problems: boolean flag indicating whether to provide the set of objects
            specified in the problem from which the trajectories have been generated

        :return: a string representing the learned PDDL model
        """

        # Format input trajectories
        os.makedirs('tmp', exist_ok=True)
        filled_traj_paths = []
        for i, traj_path in enumerate(trajectory_paths):
            filled_traj = self._preprocess_trace(traj_path)
            filled_traj_paths.append(f"tmp/{i}_traj_filled")
            with open(f"tmp/{i}_traj_filled", "w") as f:
                f.write(filled_traj)

        # Instantiate ROSAME algorithm
        partial_domain = DomainParser(Path(domain_path), partial_parsing=True).parse_domain()
        rosame = Rosame_Runner(domain_path)

        # Parse input trajectories TODO: TO BE REMOVED (if not ...)
        if not use_problems:
            allowed_observations = [TrajectoryParser(partial_domain).parse_trajectory(traj_path)
                                    for traj_path in sorted(filled_traj_paths,
                                                            key=lambda x: int(x.split('/')[-1].split('_')[0]))]
        else:
            # allowed_observations = []
            for k, traj_path in enumerate(sorted(filled_traj_paths,
                                                 key=lambda x: int(x.split('/')[-1].split('_')[0]))):
                problem_path = trajectory_paths[k].replace('trajectories', 'problems').replace('_traj', '_prob.pddl')
                problem = ProblemParser(Path(problem_path), partial_domain).parse_problem()
                rosame.add_problem(problem)
                # Learn the observation
                observation = TrajectoryParser(partial_domain, problem).parse_trajectory(traj_path)
                rosame.ground_new_trajectory()
                rosame.learn_rosame(observation)
                # allowed_observations.append(TrajectoryParser(partial_domain, problem).parse_trajectory(traj_path))

        # Remove temporary files
        shutil.rmtree('tmp')

        return rosame.rosame_to_pddl()

    def _preprocess_trace(self, traj_path: str) -> str:
        """
        Format the trajectory to make it compliant with the algorithm, by replacing
        initial state and action keywords.

        :parameter traj_path: path to the trajectory file

        :return: a string representing the formatted trajectory
        """

        with open(traj_path, 'r') as f:
            traj_str = f.read()

        # Format extra spaces
        traj_str = re.sub(r' +', ' ', traj_str)

        # Remove initial 'observation' keyword
        traj_str = traj_str.replace('(:trajectory', '(')

        # Rename `action` into `operator`
        traj_str = traj_str.replace('(:action ', '(operator: ')

        # Rename the first state into `init`
        traj_str = traj_str.replace('(:state ', '(:init ', 1)

        return traj_str



#!/usr/bin/env python3

import argparse
import random
import sys


def has_fixed_point(perm):
    return any(idx == num for idx, num in enumerate(perm))


def opposite_dir(direction):
    if direction == "in":
        return "out"
    else:
        return "in"


def create_pddl(half_segment_ids, problem_type, size, prob_no):
    objects = []
    init = []
    goal = []

    if problem_type == "simple":
        for seg in range(1, 2 * size + 1):
            for half in half_segment_ids:
                objects.append("seg%d%s - segment" % (seg, half))
                objects.append("car%d%s - car" % (seg, half))
                init.append("(on car%d%s seg%d%s)" % (seg, half, seg, half))
                goal.append("(on car%d%s seg%d%s)" % (seg, half, seg, half))
                goal.append("(analyzed car%d%s)" % (seg, half))

            if len(half_segment_ids) == 1:
                init.append("(CYCLE-2-WITH-ANALYSIS seg%d segback)" % seg)
            else:
                init.append(
                    "(CYCLE-4-WITH-ANALYSIS seg%da seg%db segbacka segbackb)"
                    % (seg, seg)
                )

        init.append("(= (total-cost) 0)")

        if len(half_segment_ids) == 1:
            objects.append("empty - car")
            objects.append("segback - segment")
            init.append("(on empty segback)")
        else:
            for half in half_segment_ids:
                objects.append("empty%s - car" % half)
                objects.append("segback%s - segment" % half)
                init.append(f"(on empty{half} segback{half})")
    else:

        if len(half_segment_ids) == 2:
            perms = {
                "in": [0] + list(range(size - 1, 0, -1)),
                "out": [0] + list(range(size - 1, 0, -1)),
            }

        for seg in range(1, size + 1):
            for direction in ["in", "out"]:
                for half in half_segment_ids:
                    objects.append("seg-%s-%d%s - segment" % (direction, seg, half))
                    objects.append("car-%s-%d%s - car" % (direction, seg, half))
                    init.append(
                        "(on car-%s-%d%s seg-%s-%d%s)"
                        % (direction, seg, half, direction, seg, half)
                    )
                    goal.append("(analyzed car-%s-%d%s)" % (direction, seg, half))

                if len(half_segment_ids) == 1:
                    for half in half_segment_ids:
                        goal.append(
                            "(on car-%s-%d%s seg-%s-%d%s)"
                            % (direction, seg, half, direction, seg, half)
                        )
                else:
                    goal.append(
                        "(on car-%s-%db seg-%s-%da)" % (direction, seg, direction, seg)
                    )
                    goal.append(
                        "(on car-%s-%da seg-%s-%db)"
                        % (
                            direction,
                            seg,
                            opposite_dir(direction),
                            perms[direction][seg - 1] + 1,
                        )
                    )

            for seg2 in range(1, size + 1):
                if len(half_segment_ids) == 1:
                    init.append("(CYCLE-2 seg-in-%d seg-out-%d)" % (seg, seg2))
                else:
                    init.append(
                        "(CYCLE-4 seg-in-%da seg-in-%db seg-out-%da seg-out-%db)"
                        % (seg, seg, seg2, seg2)
                    )

        for in_seg in range(1, problem_type[0] + 1):
            for out_seg in range(1, problem_type[1] + 1):
                if len(half_segment_ids) == 1:
                    init.append(
                        "(CYCLE-2-WITH-ANALYSIS seg-in-%d seg-out-%d)"
                        % (in_seg, out_seg)
                    )
                else:
                    init.append(
                        "(CYCLE-4-WITH-ANALYSIS seg-in-%da seg-in-%db seg-out-%da seg-out-%db)"
                        % (in_seg, in_seg, out_seg, out_seg)
                    )

        init.append("(= (total-cost) 0)")

    print(f"(define (problem scanalyzer3d-{prob_no})")
    print("  (:domain scanalyzer3d)")
    print("  (:objects")
    for objdesc in sorted(objects):
        print("    %s" % objdesc)
    print("  )")
    print("  (:init")
    for initdesc in sorted(init):
        print("    %s" % initdesc)
    print("  )")
    print("  (:goal (and")
    for goaldesc in sorted(goal):
        print("    %s" % goaldesc)
    print("  ))")
    print("  (:metric minimize (total-cost))")
    print(")")


def parse():
    parser = argparse.ArgumentParser()

    parser.add_argument("size", type=int, help="")
    parser.add_argument("segment_type", type=str, help="")
    parser.add_argument("inout", type=str, help="")
    parser.add_argument("--seed", type=int, default=1)
    parser.add_argument("--output", default=None)

    return parser.parse_args()


def main():
    args = parse()
    random.seed(args.seed)

    if args.output:
        sys.stdout = open(args.output, "w")

    prob_id = f"{args.size}-{args.segment_type}-{args.inout}"

    size = args.size
    if args.segment_type == "ab":
        half_segment_ids = ["a", "b"]
    else:
        half_segment_ids = [""]

    if args.inout == "in":
        (insize, outsize) = (size, 1)
    elif args.inout == "both":
        (insize, outsize) = (size, size)
    else:
        assert args.inout == "none"
        (insize, outsize) = (1, 1)

    create_pddl(half_segment_ids, (insize, outsize), size, prob_id)


if __name__ == "__main__":
    main()



#!/usr/bin/env python

import argparse
import random


def get_objects(args):
    str_objects = "\n"

    # -- man
    str_objects = str_objects + "     bob - man\n    "

    # -- spanner
    for i in range(1, args.spanners + 1):
        str_objects = str_objects + " spanner" + str(i)
    str_objects = str_objects + " - spanner\n    "

    # -- nuts
    for i in range(1, args.nuts + 1):
        str_objects = str_objects + " nut" + str(i)
    str_objects = str_objects + " - nut\n    "

    # -- locations
    for i in range(1, args.locations + 1):
        str_objects = str_objects + " location" + str(i)
    str_objects = str_objects + " - location\n    "

    str_objects = str_objects + " shed gate - location\n    "

    return str_objects


def get_init(args):
    str_init = "\n"
    str_init = str_init + "    (at bob shed)\n"

    for i in range(1, args.spanners + 1):
        str_init = (
            str_init
            + "    (at spanner"
            + str(i)
            + " location"
            + str(random.randint(1, args.locations))
            + ")\n"
        )
        str_init = str_init + "    (useable spanner" + str(i) + ")\n"

    for i in range(1, args.nuts + 1):
        str_init = str_init + "    (loose nut" + str(i) + ")\n"
        str_init = str_init + "    (at nut" + str(i) + " gate)\n"

    str_init = str_init + "    (link shed location1)\n"
    str_init = str_init + "    (link location" + str(args.locations) + " gate)\n"

    for i in range(1, args.locations):
        str_init = (
            str_init + "    (link location" + str(i) + " location" + str(i + 1) + ")\n"
        )

    return str_init


def get_goals(args):
    str_goal = ""
    str_goal = str_goal + "\n  (and\n"

    for i in range(1, args.nuts + 1):
        str_goal = str_goal + "   (tightened nut" + str(i) + ")\n"

    str_goal = str_goal + ")"
    return str_goal


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument("spanners", type=int, help="number of spanners (min 1)")
    parser.add_argument("nuts", type=int, help="number of nuts (min 1 and <= spanners)")
    parser.add_argument("locations", type=int, help="number of locations (min 1)")
    parser.add_argument("--seed", type=int, default=None, help="random seed (default: use time-based seed)")
    parser.add_argument("--problem-name", default="prob", help="number of locations (min 1)")
    return parser.parse_args()


def main():
    args = parse_args()

    random.seed(args.seed)

    print("(define (problem " + args.problem_name + ")")
    print(" (:domain spanner)")
    print(" (:objects " + get_objects(args) + ")")
    print(" (:init " + get_init(args) + ")")
    print(" (:goal" + get_goals(args) + "))")


main()



#! /usr/bin/env python3

"""
Openstacks generator for sequential STRIPS with negative preconditions.
"""

import argparse
import math
import random


class Task:
    def __init__(self, orders, durations):
        self.orders = orders
        self.durations = durations
        self.num_orders = len(self.orders)
        self.num_products = len(self.orders[0])

    def get_order_ids(self):
        return range(1, self.num_orders + 1)

    def get_product_ids(self):
        return range(1, self.num_products + 1)


def parse_args():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("products", type=int, help="Number of products")
    parser.add_argument("orders", type=int, help="Number of orders")
    parser.add_argument(
        "density",
        type=int,
        help="approximate percentage of products included in an order",
    )
    parser.add_argument("--domain", default="domain.pddl")
    parser.add_argument("--problem", default="problem.pddl")
    parser.add_argument("--seed", type=int, default=1)
    args = parser.parse_args()

    if not 0 <= args.density <= 100:
        raise argparse.ArgumentTypeError("density must be between 0 and 100")

    return args


def include_product_for_order(product, order, num_products, density):
    # This is based on Ioannis's code. It is designed so that products
    # are more likely to be included in orders if the distance between
    # the product and order id (before the reshuffling that happens at
    # the end) is small.
    #
    # Compared to an independent identically distributed choice for
    # each product/order pair, this should tend to favour instances
    # with small (= few stacks) solutions since it adds some locality
    # to the problem: the matrix entries (before shuffling) will tend
    # to cluster around the diagonal.

    d = abs(product - order)
    s = num_products / 6.0
    # The next line is the density function of the normal distribution
    # with mu = 0 and sigma = s for x = d.
    p = (1 / (s * (2 * math.pi) ** 0.5)) * math.exp(-(d * d) / (2 * s * s))
    # The next line is rather magical to me. If there is an
    # explanation for what this formula signifies, I don't know it.
    chance = p * density * num_products / (2.7 * (num_products / 5.0) ** 0.5)
    return random.randrange(100) <= chance


def generate_task(num_products, num_orders, density):
    products = list(range(num_products))
    orders = list(range(num_orders))
    matrix = [[0 for product in products] for order in orders]
    for order in orders:
        for product in products:
            if include_product_for_order(product, order, num_products, density):
                matrix[order][product] = 1
    for order in orders:
        if all(matrix[order][product] == 0 for product in products):
            matrix[order][random.choice(products)] = 1
    for product in products:
        if all(matrix[order][product] == 0 for order in orders):
            matrix[random.choice(orders)][product] = 1
    weights = [random.randrange(10, 100 + 1, 10) for product in products]

    random.shuffle(products)
    # Ioannis's original code, used for generating the competition
    # instances, only shuffles the orders but not the products, which
    # means that planners that process the products in numerical order
    # (or whatever) are somewhat biased. (Depending on the planner
    # and/or track, I guess this can be either positive or negative.)
    random.shuffle(orders)

    return Task(matrix, weights)


def write_problem(task, filename):
    def write(*args, **kwargs):
        print(*args, file=outfile, **kwargs)

    with open(filename, "w") as outfile:
        write("(define (problem os)")

        write("(:domain openstacks-sequencedstrips-nonADL)")

        write("(:objects")
        max_number = max(task.num_orders, task.num_products)
        for i in range(max_number + 1):
            write(f"n{i}", end=" ")
        write(" - count")
        write(")")

        write("\n(:init")

        for i in range(max_number):
            write(f"(next-count n{i} n{i + 1})", end=" ")
        write()

        write("(stacks-avail n0)")

        for i in range(1, task.num_orders + 1):
            write(f"\n(waiting o{i})")
            for j in range(1, task.num_products + 1):
                if task.orders[i - 1][j - 1] == 1:
                    write(f"(includes o{i} p{j})", end="")
            write()
        write("\n(= (total-cost) 0)")

        write(")")

        write("(:goal\n(and")
        for i in range(1, task.num_orders + 1):
            write(f"(shipped o{i})")

        write("))")
        write("(:metric minimize (total-cost))\n\n)\n")


def write_domain(task, filename):
    def write(*args, **kwargs):
        print(*args, file=outfile, **kwargs)

    with open(filename, "w") as outfile:
        write("(define (domain openstacks-sequencedstrips-nonADL)")
        write("(:requirements :typing :action-costs)")
        write("(:types order product count)")

        write("(:constants")
        for j in range(1, task.num_products + 1):
            write(f" p{j}", end="")
        write(" - product")

        for i in range(1, task.num_orders + 1):
            write(f" o{i}", end="")
        write(" - order")
        write(")\n")

        write("(:predicates")
        write("\t(includes ?o - order ?p - product)")
        write("\t(waiting ?o - order)")
        write("\t(started ?o - order)")
        write("\t(shipped ?o - order)")
        write("\t(made ?p - product)")
        write("\t(stacks-avail ?s - count)")
        write("\t(next-count ?s ?ns - count)")
        write(")\n")

        write("(:functions")
        write("(total-cost)")
        write(")\n")

        write("(:action open-new-stack")
        write(":parameters (?open ?new-open - count)")
        write(":precondition (and (stacks-avail ?open)(next-count ?open ?new-open))")
        write(":effect (and (not (stacks-avail ?open))(stacks-avail ?new-open)")
        write(" (increase (total-cost) 1))")
        write(")\n")

        # START-ORDER
        write("(:action start-order")
        write(":parameters (?o - order ?avail ?new-avail - count)")
        write(":precondition ")
        write("(and (waiting ?o)")
        write("(stacks-avail ?avail)(next-count ?new-avail ?avail))")
        write(":effect (and (not (waiting ?o))(started ?o)")
        write("(not (stacks-avail ?avail))(stacks-avail ?new-avail))")
        write(")\n")

        # MAKE-PRODUCT
        for j in task.get_product_ids():
            write(f"(:action make-product-p{j}")
            write(":parameters ()")
            write(":precondition ")

            write(f"(and (not (made p{j}))")

            for i in task.get_order_ids():
                if task.orders[i - 1][j - 1] == 1:
                    write(f"(started o{i})")
            write(")")

            write(f":effect (and (made p{j}))")
            write(")\n")

        # SHIP-ORDER
        for i in task.get_order_ids():
            write(f"(:action ship-order-o{i}")
            write(":parameters (?avail ?new-avail - count)")
            write(f":precondition (and (started o{i})", end="")
            for j in task.get_product_ids():
                if task.orders[i - 1][j - 1] == 1:
                    write(f"(made p{j})", end="")
            write("(stacks-avail ?avail)(next-count ?avail ?new-avail))")

            write(f":effect (and (not (started o{i}))(shipped o{i})", end="")
            write("(not (stacks-avail ?avail))(stacks-avail ?new-avail)))\n")

        write(")\n")


def main():
    args = parse_args()
    random.seed(args.seed)
    task = generate_task(args.products, args.orders, args.density)
    write_problem(task, args.problem)
    write_domain(task, args.domain)


main()



import re
from dataclasses import dataclass
from typing import List

from unified_planning.model import UPState
from unified_planning.plans import ActionInstance


@dataclass
class Trajectory:

    states: List[UPState]
    actions: List[ActionInstance]

    def __str__(self):

        def format_pddl_string(literals: List[str]) -> str:
            return ' '.join([f"({l.replace('(', ' ').replace(')', '').replace(',', ' ')})"
                             for l in sorted(literals)])

        init_state = re.findall(r'(\w+(?:\([^\)]*\))?)\s*:\s*true', str(self.states[0]))  # positive literals
        trajectory_str = f"(:state {format_pddl_string(init_state)})"
        for i in range(len(self.actions)):
            a = f"({repr(self.actions[i]).replace('(', ' ').replace(')', '').replace(',', '')})"
            s = re.findall(r'(\w+(?:\([^\)]*\))?)\s*:\s*true', str(self.states[i + 1]))  # positive literals
            trajectory_str += f"\n\n(:action {a})"
            trajectory_str += f"\n\n(:state {format_pddl_string(s)})"

        trajectory_str = f"(:trajectory\n\n{trajectory_str}\n\n)".replace('  ', ' ')
        return trajectory_str

    def __repr__(self):
        return str(self)

    def write(self, file_path):
        with open(file_path, 'w') as f:
            f.write(str(self))



from collections import defaultdict
import sys

import itertools
from typing import List, Union, Optional, Dict


from pddl_plus_parser.models import Domain, Observation, State, ActionCall, PDDLFunction, Predicate, GroundedPredicate, create_type_hierarchy_graph

from pddl_plus_parser.lisp_parsers import ProblemParser, TrajectoryParser, DomainParser

from models.rosame import Action_Schema,Domain_Model,Predicate,Type
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch.utils.data import TensorDataset, DataLoader, random_split
import numpy as np
import networkx as nx
from itertools import permutations


def parse_state(state, partial_domain):
    state_predicates = defaultdict(set)
    for predicate in state:
        predicate_lst = predicate.split()
        if predicate_lst[0] in partial_domain.predicates:
            lifted_predicate = partial_domain.predicates[predicate_lst[0]]
            grounded_predicate = parse_grounded_predicate(predicate_lst, lifted_predicate)
            state_predicates[lifted_predicate.untyped_representation].add(grounded_predicate)
            continue
    return State(predicates=state_predicates, fluents={})

def parse_grounded_predicate(grounded_predicate_ast: List[str],
                                 lifted_predicate: Predicate) -> GroundedPredicate:

    predicate_name = lifted_predicate.name
    predicate_signature_items = grounded_predicate_ast[1:]

    if len(predicate_signature_items) != len(lifted_predicate.signature):
        raise ValueError(
            f"Received illegal grounded predicate with mismatching signature - {grounded_predicate_ast}")

    object_mapping = {
        parameter_name: object_name for object_name, parameter_name in
        zip(predicate_signature_items, lifted_predicate.signature)
    }

    return GroundedPredicate(name=predicate_name, signature=lifted_predicate.signature,
                             object_mapping=object_mapping, is_positive=True)

class Rosame_Runner:
    '''
    A class for learning algorithms that prepares the data and learn an action model from generated traces.x
    '''
    def __init__(self, domain_file, problem = None, traces = None, threshold = 1, epsilon = 0):
        self.traces = traces
        self.domain = DomainParser(domain_file,partial_parsing=True).parse_domain()
        self.problem = problem
        self.trajectoy = TrajectoryParser(self.domain)
        self.grounded_actions = {}
        self.objects = {}
        self.type_hierarchy_graph = None
        self.rosame = None
        self.epsilon = epsilon
        self.threshold = threshold
        self.optimizer = None
        self.opt_flag = True

    def add_problem(self,problem):
        self.problem = problem
        self.rosame = self.prepare_rosame()

        # if self.opt_flag:
        #     parameters = []
        #     for schema in self.rosame.action_schemas:
        #         parameters.append({'params': schema.parameters(), 'lr': 1e-3})
        #     self.optimizer = optim.Adam(parameters)
        #     self.opt_flag = False


    def find_root_nodes(self,G: nx.DiGraph) -> str:
        """
        Finds the root nodes of the directed graph (nodes with in-degree 0).
        """
        return [node for node in G.nodes if G.in_degree(node) == 0][0]

    def _prepare_types(self,domain_types):
        """
        Creates Type: parent representation for the types of the problem
        """
        self.type_hierarchy_graph = create_type_hierarchy_graph(domain_types)
        root = self.find_root_nodes(self.type_hierarchy_graph)
        hierarchy_with_parents = [{'name': child,'parent': parent} for parent, child in nx.bfs_edges(self.type_hierarchy_graph, root)]
        return [{'name': 'object','parent': None}] + hierarchy_with_parents

    def _get_types_count(self,container):
        """
        A map that countes for each predicate/action what are the types used and how many of them
        """
        params = defaultdict(int)
        for _, types in container.signature.items():
            params[types.name] += 1
        return params

    def _prepare_predicates(self,domain_predicates):
        """
        Creates predicate map that has name and dict of type: count
        """
        predicates_lst = []
        for name, predicate in domain_predicates.items():
            predicate_dict = {}
            predicate_dict["name"] = name
            predicate_dict["params"] = self._get_types_count(predicate)
            predicates_lst.append(predicate_dict)
        return predicates_lst


    def _prepare_action_schema(self,domain_actions):
        """
        Creates action map that has name and dict of type: count
        """
        actions_lst = []
        for name, action in domain_actions.items():
            action_dict = {}
            action_dict["name"] = name
            action_dict["params"] = self._get_types_count(action)
            actions_lst.append(action_dict)
        return actions_lst

    def get_domain_model(self,info_dict):
        """
        The function loads domain model from the information of actions,predicates and types
        """
        domain_model = Domain_Model.create_from_json(info_dict,"cpu")
        domain_model.ground_from_dict(self.objects)
        return domain_model


    def get_objects(self,objects):
        """
        function to map the objects by type
        """
        object_dict = defaultdict(list)
        for name,t in objects.items():
            object_dict[t.type.name].append(name)
        if len(self.domain.constants) > 0:
            for name, t in self.domain.constants.items():
                object_dict[t.type.name].append(name)
        return object_dict

    def check_rosame_action(self, action: str):
        if action in self.rosame.actions:
            return action

        parts = action.split(" ", maxsplit=1)
        predicate_name, predicate_params = parts[0], parts[1].split(" ")
        predicate_params_types = predicate_params[::2]
        predicate_params_values = predicate_params[1::2]

        # Find the first matching proposition with the same predicate name
        rosame_predicate = next((p for p in self.rosame.actions if p.startswith(predicate_name)), None)

        if not rosame_predicate:
            return action  # No matching predicate found

        rosame_parts = rosame_predicate.split(" ")
        rosame_types = rosame_parts[1::2]

        if rosame_types == predicate_params_types:
            return action  # Exact match, return as is

        # Replace types with their parent if needed
        for i, (param_type, rosame_type) in enumerate(zip(predicate_params_types, rosame_types)):
            if param_type != rosame_type and rosame_type in nx.ancestors(self.type_hierarchy_graph,param_type): ## for direct parent we can use self.type_hierarchy_graph.predecessors(param_type)
                predicate_params_types[i] = rosame_type

        return f"{rosame_parts[0]} " + " ".join(
            f"{t} {v}" for t, v in zip(predicate_params_types, predicate_params_values))


    def check_rosame_predicate(self, predicate: str):
        if predicate in self.rosame.propositions:
            return predicate

        parts = predicate.split(" ", maxsplit=1)
        predicate_name, predicate_params = parts[0], parts[1].split(" ")
        predicate_params_types = predicate_params[::2]
        predicate_params_values = predicate_params[1::2]

        # Find the first matching proposition with the same predicate name
        rosame_predicate = next((p for p in self.rosame.propositions if p.startswith(predicate_name)), None)

        if not rosame_predicate:
            return predicate  # No matching predicate found

        rosame_parts = rosame_predicate.split(" ")
        rosame_types = rosame_parts[1::2]

        if rosame_types == predicate_params_types:
            return predicate  # Exact match, return as is

        # Replace types with their parent if needed
        for i, (param_type, rosame_type) in enumerate(zip(predicate_params_types, rosame_types)):
            if param_type != rosame_type and rosame_type in nx.ancestors(self.type_hierarchy_graph,param_type): ## for direct parent we can use self.type_hierarchy_graph.predecessors(param_type)
                predicate_params_types[i] = rosame_type

        return f"{predicate_name} " + " ".join(
            f"{t} {v}" for t, v in zip(predicate_params_types, predicate_params_values))

    def prepare_rosame(self):
        """
            Prepares and constructs a ROSAME-compatible domain model based on the provided domain definitions.

            This function follows a structured process to generate a complete domain model:

            1. Types: Constructs a hierarchical representation of types.
            2. Predicates: Defines logical predicates using.
            3. Action Schemas: Creates action schemas, which define the available actions and their required parameters.
            4. Objects: Extracts objects from the problem definition and stores them.
            5. Domain Model: Combines all components into a structured `Domain_Model` instance.

            Returns:
                Domain_Model: A structured domain model instance containing:
                    - Types
                    - Predicates
                    - Action Schemas
        """
        types = self._prepare_types(self.domain.types)
        predicates = self._prepare_predicates(self.domain.predicates)
        actions = self._prepare_action_schema(self.domain.actions)
        self.objects = self.get_objects(self.problem.objects)
        model = self.get_domain_model({"types":types,"predicates":predicates,"action_schemas":actions})

        return model

    def ground_new_trajectory(self):
        self.objects = self.get_objects(self.problem.objects)
        self.rosame.ground_from_dict(self.objects)

    def generate_noise(self,trajectory):
        if self.epsilon == 0:
            return trajectory
        noisy_trajectory = []
        np.random.seed(123)
        for f in trajectory:
            if f == 0:
                f_noise = np.random.uniform(0, self.epsilon)
            else:
                f_noise = 1 - np.random.uniform(0, self.epsilon)
            noisy_trajectory.append(f_noise)
        return noisy_trajectory

    def check_predicate(self, predicate_str: str):
        if predicate_str.rstrip() in self.rosame.propositions:
            return predicate_str.rstrip()

        parts = predicate_str.split()
        action_name = parts[0]
        arguments = parts[1:]

        # Generate all permutations of the arguments
        for perm in permutations(arguments):
            candidate = f"{action_name} {' '.join(perm)}"
            if candidate in self.rosame.propositions:
                return candidate

    def check_action(self, action_str: str):
        if action_str.rstrip() in self.rosame.actions:
            return self.rosame.actions[action_str.rstrip()]

        parts = action_str.split()
        action_name = parts[0]
        arguments = parts[1:]

        # Generate all permutations of the arguments
        for perm in permutations(arguments):
            candidate = f"{action_name} {' '.join(perm)}"
            if candidate in self.rosame.actions:
                return self.rosame.actions[candidate]

    def prepare_rosame_data(self,observation):
        """
        Prepares structured data from traces to be used within the ROSAME framework, which the data should be encoded to True/False (0  and 1)
        using the triples that are created using macq every fluent is true takes 1 else 0
        returns:
        - encoded pre_state
        - encoded next_state
        - encoded action
        """
        steps_action = []
        steps_state1, steps_state2 = [], []
        for component in observation.components:
            action_lst = component.grounded_action_call.__str__()[1:-1].split()
            if len(action_lst) != len(set(action_lst)):
                continue
            next_state, pre_state = [], []
            steps_action.append(self.check_action(component.grounded_action_call.__str__()[1:-1]))

            for _, val in component.next_state.state_predicates.items():
                for pred in val:
                    next_state.append(self.check_predicate(pred.untyped_representation[1:-1]))

            for _, val in component.previous_state.state_predicates.items():
                for pred in val:
                    pre_state.append(self.check_predicate(pred.untyped_representation[1:-1]))

            state1 = [1 if p in pre_state else 0 for p in self.rosame.propositions]
            state2 = [1 if p in next_state else 0 for p in self.rosame.propositions]
            steps_state1.append(state1)
            steps_state2.append(state2)

        return steps_state1, steps_action, steps_state2

    def learn_rosame(self, observation,epochs=100):
        """Learns an action model using ROSAME learning."""

        steps_state1, steps_action, steps_state2 = self.prepare_rosame_data(observation)

        steps_state1_tensor = torch.tensor(np.array(steps_state1)).float()
        steps_action_tensor = torch.tensor(np.array(steps_action))
        steps_state2_tensor = torch.tensor(np.array(steps_state2)).float()

        batch_sz = 1000
        dataset = TensorDataset(steps_state1_tensor, steps_action_tensor, steps_state2_tensor)
        dataloader = DataLoader(dataset, batch_size=batch_sz, shuffle=False)

        parameters = [] #TODO: maybe have to be initialized only one time per learning process(?)
        for schema in self.rosame.action_schemas:
            parameters.append({'params': schema.parameters(), 'lr': 1e-3})
        optimizer = optim.Adam(parameters)

        for epoch in range(epochs):
            loss_final = 0.0
            for i, (state_1, executed_actions, state_2) in enumerate(dataloader):
                optimizer.zero_grad()
                precon, addeff, deleff = self.rosame.build(executed_actions)
                preds = state_1 * (1 - deleff) + (1 - state_1) * addeff
                loss = F.mse_loss(preds, state_2, reduction='sum')
                validity_constraint = (1 - state_1) * (precon)
                loss += F.mse_loss(validity_constraint,
                                   torch.zeros(validity_constraint.shape, dtype=validity_constraint.dtype),
                                   reduction='sum')
                #     loss += model.constraint_loss()
                loss += 0.2 * F.mse_loss(precon, torch.ones(precon.shape, dtype=precon.dtype), reduction='sum')
                loss.backward()
                optimizer.step()
                loss_final += loss.item() / batch_sz
            if epoch % 10 == 0:
                print('Epoch {} RESULTS: Average loss: {:.10f}'.format(epoch, loss_final))


    def action_params_pddl(self, params):
        """
        a help function that helps to create action string for pddl
        """
        param_str = " ".join(f"?{param} - {type_.name}" for type_, p in params.items() for param in p)
        return f"({param_str})\n"

    def format_conditions(self, conditions, negate=False):
        formatted = []
        for cond in conditions:
            parts = cond.split()
            if len(parts) > 1:
                args = " ".join(f"?{parts[i]}" for i in range(1, len(parts)))
                if negate:
                    formatted.append(f" (not ({parts[0]} {args}))")
                else:
                    formatted.append(f"({parts[0]} {args})")
            else:
                if negate:
                    formatted.append(f" (not ({parts[0]}))")
                else:
                    formatted.append(f"({parts[0]})")
        return " ".join(formatted)

    def precondition_pddl(self, pre):
        return f"(and {self.format_conditions(pre)})\n"

    def effects_pddl(self, add, delete):
        return f"(and {self.format_conditions(add)}{self.format_conditions(delete, negate=True)}))\n\n"

    def _types_to_pddl(self,types):
        parent_child_map = defaultdict(list)
        for type_name, type_obj in types.items():
            if type_name == "object":
                continue

            parent_child_map[type_obj.parent.name].append(type_name)

        types_strs = []
        for parent_type, children_types in parent_child_map.items():
            types_strs.append(f"\t{' '.join(children_types)} - {parent_type}")

        return "\n".join(types_strs)

    def _signature_to_pddl(self,action) -> str:
        """Converts the action's signature to the PDDL format.

        :return: the PDDL format of the signature.
        """
        signature_str_items = []
        for parameter_name, parameter_type in self.domain.actions[action.name].signature.items():
            signature_str_items.append(f"{parameter_name} - {str(parameter_type)}")

        signature_str = " ".join(signature_str_items)
        return f"({signature_str})"

    def get_params_names(self,action):
        param_types = defaultdict(list)
        for parameter_name, parameter_type in action.signature.items():
            param_types[str(parameter_type)].append(parameter_name[1:])
        return param_types

    def to_pddl_action(self, action):
        """Returns the PDDL string representation of the action.
        :return: the PDDL string representing the action.
        """
        params = self.get_params_names(self.domain.actions[action.name])
        precondition, add_eff, delete_eff, param = action.pretty_print(params) # TODO: Fix names
        action_string = (
            f"(:action {action.name}\n"
            f"\t:parameters {self._signature_to_pddl(action)}\n" # TODO: Fix order
            f"\t:precondition {self.precondition_pddl(precondition)}\n"
            f"\t:effect {self.effects_pddl(add_eff,delete_eff)}"
        )
        formatted_string = "\n".join([line for line in action_string.split("\n") if line.strip()])
        return f"{formatted_string}\n"

    def _constants_to_pddl(self,domain) -> str:
        """Converts the constants to a PDDL string.
        :return: the PDDL string representing the constants.
        """
        same_type_constant = defaultdict(list)
        for const_name, constant in domain.constants.items():
            if const_name == "object":
                continue

            same_type_constant[constant.type.name].append(const_name)

        types_strs = []
        for constant_type_name, constant_objects in same_type_constant.items():
            types_strs.append(f"\t{' '.join(constant_objects)} - {constant_type_name}")

        return "\n".join(types_strs)

    def rosame_to_pddl(self):
        predicates = "\n\t".join([str(p) for p in self.domain.predicates.values()])
        predicates_str = f"(:predicates {predicates}\n)\n\n" if len(self.domain.predicates) > 0 else ""
        functions = "\n\t".join([str(p) for p in self.domain.functions.values()])
        functions_str = f"(:functions {functions}\n)\n\n" if len(self.domain.functions) > 0 else ""
        types = f"(:types {self._types_to_pddl(self.domain.types)}\n)\n\n" if len(self.domain.types) > 0 else ""
        constants = f"(:constants {self._constants_to_pddl(self.domain)}\n)\n\n" if len(self.domain.constants) > 0 else ""
        actions = "\n".join(self.to_pddl_action(action) for action in self.rosame.action_schemas)

        return (
            f"(define (domain {self.domain.name})\n"
            f"(:requirements {' '.join(self.domain.requirements)})\n"
            f"{types}"
            f"{constants}"
            f"{predicates_str}"
            f"{functions_str}"
            f"{actions}\n)"
        )
    def export_rosame_domain(self,rosame_path):
        with open(rosame_path, "w") as f:
            f.write(self.rosame_to_pddl())

    def export_sam_domain(self,learned_model,path):
        with open(path, "w") as f:
            f.write(learned_model.to_pddl())






import os
import re
import shutil
from dataclasses import dataclass
from pathlib import Path
from typing import List, ClassVar, OrderedDict

from pddl_plus_parser.lisp_parsers import DomainParser, TrajectoryParser
from sam_learning.learners import SAMLearner

from amlgym.algorithms.OfflineAlgorithmAdapter import OfflineAlgorithmAdapter


@dataclass
class SAM(OfflineAlgorithmAdapter):
    """
    Adapter class for running the SAM algorithm: "Safe Learning of Lifted Action Models",
    B. Juba and H. S. Le, and R. Stern, Proceedings of the 18th International Conference
    on Principles of Knowledge Representation and Reasoning, 2021.
    https://proceedings.kr.org/2021/36/

    Example:
        .. code-block:: python

            from amlgym.algorithms import get_algorithm
            sam = get_algorithm('SAM')
            model = sam.learn('path/to/domain.pddl', ['path/to/trace0', 'path/to/trace1'])
            print(model)

    """
    _reference: ClassVar[OrderedDict[str, str]] = {
        'Authors': "B. Juba and H. S. Le, and R. Stern",
        'Title': "Safe Learning of Lifted Action Models",
        'Venue': "International Conference on Principles of Knowledge Representation and Reasoning",
        'Year': 2021,
        'URL': "https://proceedings.kr.org/2021/36/",
    }

    def learn(self,
              domain_path: str,
              trajectory_paths: List[str]) -> str:
        """
        Learns a PDDL action model from:
         (i)    a (possibly empty) input model which is required to specify the predicates and operators signature;
         (ii)   a list of trajectory file paths.

        :parameter domain_path: input PDDL domain file path
        :parameter trajectory_paths: list of trajectory file paths

        :return: a string representing the learned PDDL model
        """

        # Format input trajectories
        os.makedirs('tmp', exist_ok=True)
        filled_traj_paths = []
        for i, traj_path in enumerate(trajectory_paths):
            filled_traj = self._preprocess_trace(traj_path)
            filled_traj_paths.append(f"tmp/{i}_traj_filled")
            with open(f"tmp/{i}_traj_filled", "w") as f:
                f.write(filled_traj)

        # Instantiate SAM algorithm
        partial_domain = DomainParser(Path(domain_path), partial_parsing=True).parse_domain()
        sam = SAMLearner(partial_domain=partial_domain)
        allowed_observations = []
        for k, traj_path in enumerate(sorted(filled_traj_paths,
                                             key=lambda x: int(x.split('/')[-1].split('_')[0]))):
            allowed_observations.append(TrajectoryParser(partial_domain).parse_trajectory(traj_path))

        # Learn action model
        learned_model, learning_report = sam.learn_action_model(allowed_observations)

        # Remove temporary files
        shutil.rmtree('tmp')

        return learned_model.to_pddl()

    def _preprocess_trace(self, traj_path: str) -> str:
        """
        Format the trajectory to make it compliant with the algorithm, by replacing
        initial state and action keywords.

        :parameter traj_path: path to the trajectory file

        :return: a string representing the formatted trajectory
        """

        with open(traj_path, 'r') as f:
            traj_str = f.read()

        # Format extra spaces
        traj_str = re.sub(r' +', ' ', traj_str)

        # Remove initial 'observation' keyword
        traj_str = traj_str.replace('(:trajectory', '(')

        # Rename `action` into `operator`
        traj_str = traj_str.replace('(:action ', '(operator: ')

        # Rename the first state into `init`
        traj_str = traj_str.replace('(:state ', '(:init ', 1)

        return traj_str



# Add current project to sys path
import os
import sys
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, parent_dir)

from .gen_problems import *  # do not remove

from datetime import datetime
import yaml
import re
import contextlib
import logging
import random
import shutil
import numpy as np
import unified_planning
from alive_progress import alive_bar
from unified_planning.io import PDDLReader
from unified_planning.shortcuts import OneshotPlanner


if __name__ == '__main__':

    MAX_PLANNING_TIME = 60
    PLANS_MIN_LEN = 5
    DOWNWARD_SEARCH_CFG = 'let(hff,ff(),let(hcea,cea(),lazy_greedy([hff,hcea],preferred=[hff,hcea])))'
    PLANNER_CFG = {  # heuristic planning
        'name': 'fast-downward',
        'params': dict(
            fast_downward_search_config=DOWNWARD_SEARCH_CFG,
            fast_downward_search_time_limit=f"{MAX_PLANNING_TIME}s"
        )}
    logging.basicConfig(
        # filename='out.log',
        level=logging.DEBUG
    )
    GEN_DIR = "pddl-generators"
    BENCHMARK_DIR = "benchmarks"
    DOMAINS_DIR = "domains"
    PROB_DIR = "problems/solving"
    DOM_CFG = f"{BENCHMARK_DIR}/problems_solving.yaml"

    # Trace CPU time
    run_start = datetime.now()

    # Instantiate a PDDL problem reader
    reader = PDDLReader()

    # Disable printing of planning engine credits to avoid overloading stdout
    unified_planning.shortcuts.get_environment().credits_stream = None

    # Read domain configs
    with open(f"../{DOM_CFG}") as f:
        cfg = yaml.safe_load(f)
        seed = cfg['SEED']
        domains = cfg['domains']

    to_be_avoided = []
    domains = {k: v for k, v in domains.items() if k not in to_be_avoided}

    for domain in domains:

        # Set random seed for reproducibility
        np.random.seed(seed)
        random.seed(seed)

        # Trace CPU time
        domain_run_start = datetime.now()

        # Clean domain problems directory
        if os.path.exists(f"../{BENCHMARK_DIR}/{PROB_DIR}/{domain}"):
            shutil.rmtree(f"../{BENCHMARK_DIR}/{PROB_DIR}/{domain}")
        os.makedirs(os.path.join(f"../{BENCHMARK_DIR}/{PROB_DIR}/{domain}"))

        with alive_bar(len(domains[domain]),
                       title=f'Processing domain {domain}',
                       length=20,
                       bar='halloween') as bar:
            # For every domain problem kwargs
            for i, kwargs in enumerate(domains[domain]):

                plan = None

                while (plan is None) or len(plan.actions) < PLANS_MIN_LEN:
                    # Generate a problem
                    logging.debug(f"Generating a new problem")
                    kwargs['seed'] = np.random.randint(1, 1000)
                    generate_prob = getattr(sys.modules[__name__], f'problem_{domain}')
                    problem_str = generate_prob(**kwargs)
                    # Fix hyphens to avoid issues with unified-planning parsing
                    problem_str = re.sub(r'(?<=\w)-(?=\w)', '_', problem_str)

                    # Write the problem string to a file
                    problems_dir = f"../{BENCHMARK_DIR}/{PROB_DIR}/{domain}"
                    problem_file = f'{problems_dir}/{len(os.listdir(problems_dir))}_{domain}_prob.pddl'
                    with open(problem_file, 'w') as f:
                        f.write(problem_str.lower())

                    # Parse the problem in unified-planning
                    domain_file = f'../{BENCHMARK_DIR}/{DOMAINS_DIR}/{domain}.pddl'
                    problem = reader.parse_problem(domain_file, problem_file)

                    # Solve the problem
                    logging.debug("Computing a new plan...")
                    with contextlib.redirect_stdout(open(os.devnull, 'w')):
                        with OneshotPlanner(
                                problem_kind=problem.kind,
                                **PLANNER_CFG
                        ) as planner:
                            result = planner.solve(problem, timeout=MAX_PLANNING_TIME)
                            plan = result.plan

                    if (plan is None) or len(plan.actions) < PLANS_MIN_LEN:
                        if plan is None:
                            logging.debug(f"Failed to generate a plan. Retrying...")
                        elif len(plan.actions) < PLANS_MIN_LEN:
                            logging.debug(f"Plan is not sufficiently long. Plan length is {len(plan.actions)}"
                                          f", minimum length is {PLANS_MIN_LEN}. Retrying...")
                        else:
                            raise Exception('Something went wrong during plan generation.')
                        plan = None
                        os.remove(problem_file)

                bar()  # update progress bar

    logging.info(f'Total CPU time (s): {(datetime.now() - run_start).seconds}')



import torch
from torch.utils.data import Dataset
from torchvision import transforms
import torchvision
import random


class RearrangeColumn(object):
    def __init__(self, column_num):
        self.column_num = column_num

    def __call__(self, img):
        idx = torch.randperm(self.column_num)
        return torch.cat((img[:, [0]], img[:, 1:, idx]), 1)


class RearrangeBalls(object):
    def __init__(self, column_num):
        self.column_num = column_num

    def __call__(self, img):
        # img is steps * row * column * 28 * 28
        idx1 = torch.randperm(self.column_num)
        idx2 = torch.randperm(self.column_num)
        return torch.cat(
            (
                img[:, [0]],
                img[:, [1], idx1].unsqueeze(1),
                img[:, [2]],
                img[:, [3], idx2].unsqueeze(1),
            ),
            1,
        )


class RearrangeItems(object):
    def __call__(self, img):
        # img is steps * row * column * 3 * 28 * 28
        indices = [
            [
                [
                    (r, c)
                    for r in range(i * 3, i * 3 + 3)
                    for c in range(j * 3, j * 3 + 3)
                ]
                for j in range(2)
            ]
            for i in range(2)
        ]
        for i in range(2):
            for j in range(2):
                random.shuffle(indices[i][j])
        rows = []
        columns = []
        for r in range(6):
            for c in range(6):
                idx = indices[int(r / 3)][int(c / 3)].pop(0)
                rows.append(idx[0])
                columns.append(idx[1])
        return img[:, rows, columns, :, :, :].unflatten(1, (6, 6))


class GridDataset(Dataset):
    def __init__(self, dataset_path, step_length, transforms=None):
        self.dataset_path = dataset_path
        self.step_length = step_length
        self.transforms = transforms

        with open(f"{dataset_path}/features_img.pt", "rb") as f:
            self.images = torch.load(f)
            if self.images.dim() == 6:
                self.images = self.images.unsqueeze(4).float()
            else:
                self.images = self.images.float()
        with open(f"{dataset_path}/labels.pt", "rb") as f:
            self.labels = torch.load(f)
        with open(f"{dataset_path}/actions.pt", "rb") as f:
            self.actions = torch.load(f)

        self.images = self.images[:, :self.step_length]
        self.labels = self.labels[:, :self.step_length+1]
        self.actions = self.actions[:, :self.step_length]

    def __getitem__(self, index):
        img = self.images[index]
        label = self.labels[index]
        action = self.actions[index]

        if self.transforms is not None:
            img = self.transforms(img)
        return img, label, action

    def __len__(self):
        return len(self.images)


class SynthDataset(Dataset):
    def __init__(self, dataset_path, step_length, skip=1, transforms=None):
        self.dataset_path = dataset_path
        self.step_length = step_length
        self.transforms = transforms

        if skip == "break_symmetry":
            # Break symmetry based on whether the trace length is even or odd
            # Always skip at least one state
            self.skip = 3 - self.step_length % 2
        else:
            self.skip = skip

        with open(f"{dataset_path}/labels.pt", "rb") as f:
            self.labels = torch.load(f)
        with open(f"{dataset_path}/actions.pt", "rb") as f:
            self.actions = torch.load(f)

    def __getname__(self, idx):
        return f"{self.dataset_path}/{idx}.png"

    def __len__(self):
        return int(self.actions.shape[0] / (self.step_length + self.skip))

    def __getitem__(self, idx):
        # For some reason we failed to save the first 10 images
        starting_idx = idx * (self.step_length + self.skip)
        images = [
            torchvision.io.read_image(
                self.__getname__(starting_idx + i),
                mode=torchvision.io.ImageReadMode.RGB,
            )
            for i in range(self.step_length)
        ]
        images = torch.stack(images, dim=0)
        images = images.float()
        labels = self.labels[starting_idx : starting_idx + self.step_length + 1]
        actions = self.actions[starting_idx : starting_idx + self.step_length]

        if self.transforms:
            images = self.transforms(images)

        return images, labels, actions


# Add current project to sys path
import os
import sys
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
sys.path.insert(0, parent_dir)

from .gen_problems import *  # do not remove

from datetime import datetime
import yaml
import re
import contextlib
import logging
import random
import shutil
import numpy as np
import unified_planning
from alive_progress import alive_bar
from unified_planning.io import PDDLReader
from unified_planning.model import Problem, UPState
from unified_planning.plans import ActionInstance
from unified_planning.shortcuts import OneshotPlanner, SequentialSimulator

from amlgym.modeling.trajectory import Trajectory

from tarski.io import PDDLReader as tarskiPDDLReader
from tarski.grounding import LPGroundingStrategy


def replan(problem: Problem,
           current_state: UPState,
           action_instance: ActionInstance) -> any:
    """
    Check action execution does not make the problem unsolvable by simulating the action and computing
    a new plan
    :param problem: solvable problem
    :param action: action to be executed
    :return: new plan (if any)
    """
    # Update the problem initial state
    problem = problem.clone()
    for fluent in problem.initial_values:
        value = current_state.get_value(fluent)
        problem.set_initial_value(fluent, value)

    # Simulate action execution
    with SequentialSimulator(problem=problem) as simulator:
        current_state = simulator.apply(current_state, action_instance)

    # Update the problem state
    problem = problem.clone()
    for fluent in problem.initial_values:
        value = current_state.get_value(fluent)
        problem.set_initial_value(fluent, value)

    # Check a plan still exists
    logging.debug(f"Checking random action {action_instance} preserves solvability.")
    with contextlib.redirect_stdout(open(os.devnull, 'w')):
        with OneshotPlanner(
                problem_kind=problem.kind,
                **PLANNER_CFG,
        ) as planner:
            result = planner.solve(problem, timeout=MAX_REPLANNING_TIME)
            plan = result.plan

    return plan


def generate_traj(
        problem: Problem,
        randomness: float = 0.2) -> Trajectory:

    with SequentialSimulator(problem=problem) as simulator:

        current_state = simulator.get_initial_state()
        states = [current_state]  # init trajectory states
        actions = []  # init trajectory actions
        plan = None

        # Ground actions with tarski since unified-planning (1.2.0) grounder is inefficient
        reader = tarskiPDDLReader(raise_on_error=True)
        reader.parse_domain(domain_file)
        reader.parse_instance(problem_file)
        grounder = LPGroundingStrategy(reader.problem)
        ground_actions = grounder.ground_actions()

        while len(states) < TRAJ_LEN_MAX:

            if plan is None:
                # Update the problem initial state
                logging.debug(f"Updating problem state")
                problem = problem.clone()
                for fluent in problem.initial_values:
                    value = current_state.get_value(fluent)
                    problem.set_initial_value(fluent, value)

                logging.debug("Computing a new plan...")
                with contextlib.redirect_stdout(open(os.devnull, 'w')):
                    with OneshotPlanner(
                            problem_kind=problem.kind,
                            **PLANNER_CFG
                    ) as planner:
                        result = planner.solve(problem, timeout=MAX_PLANNING_TIME)
                        plan = result.plan

            # Problem unsolvable
            if plan is None:
                if result.status.name == 'TIMEOUT':
                    logging.debug(f"Planning timout reached ({MAX_PLANNING_TIME}s).")
                    break
                elif result.status.name == 'UNSOLVABLE_INCOMPLETELY':
                    logging.debug(f"Planning unsolvable.")
                    break
                else:
                    raise Exception(f"Planning exited with status: {result.status.name}")

            for action_instance in plan.actions:

                # Possibly execute a random action and replan
                if random.random() < randomness:
                    logging.debug(f"Sampling a random action...")

                    # applicable_actions = list(simulator.get_applicable_actions(current_state))
                    applicable_actions = [(problem.action(k.lower()), [problem.object(o.lower()) for o in objs])
                                          for k, params in ground_actions.items()
                                          for objs in params
                                          if simulator._is_applicable(current_state,
                                                                      problem.action(k.lower()),
                                                                      [problem.object(o.lower()) for o in objs])]

                    applicable_actions = sorted(applicable_actions, key=lambda x: f"{x[0]} - {x[1]}")  # reproducibility
                    action, params = random.choices(applicable_actions)[0]
                    action_instance = ActionInstance(action, params)
                    logging.debug(f"Random action sampled.")

                    # Check random action does not make the problem unfeasible
                    trial = 1
                    plan = replan(problem, current_state, action_instance)
                    # while not check_feasibility(problem, current_state, action_instance):
                    while plan is None:
                        trial += 1
                        logging.debug(f"Random action {action_instance} makes the problem unsolvable."
                                      f" Newly sampling a random action.")

                        # applicable_actions = list(simulator.get_applicable_actions(current_state))
                        applicable_actions = [(problem.action(k.lower()), [problem.object(o.lower()) for o in objs])
                                              for k, params in ground_actions.items()
                                              for objs in params
                                              if simulator._is_applicable(current_state,
                                                                          problem.action(k.lower()),
                                                                          [problem.object(o.lower()) for o in objs])]

                        applicable_actions = sorted(applicable_actions, key=lambda x: f"{x[0]} - {x[1]}")  # reproducibility
                        action, params = random.choices(applicable_actions)[0]
                        action_instance = ActionInstance(action, params)
                        plan = replan(problem, current_state, action_instance)
                        if trial >= MAX_RANDOM_TRIALS and plan is None:
                            break

                    if trial >= MAX_RANDOM_TRIALS and plan is None:
                        logging.debug(f"Maximum number of random action trials reached."
                                      f" Avoiding random action execution.")
                        break

                    logging.debug(f"Simulating random action {action_instance}.")
                    current_state = simulator.apply(current_state, action_instance)
                    states.append(current_state)
                    actions.append(action_instance)
                    break

                logging.debug(f"Simulating action {action_instance}.")
                current_state = simulator.apply(current_state, action_instance)
                actions.append(action_instance)

                if current_state is None:
                    raise Exception(f"Error in applying: {action_instance}")

                states.append(current_state)

                if len(states) == TRAJ_LEN_MAX:
                    break

            if plan is not None and (len(plan.actions) == 0 or action_instance == plan.actions[-1]):
                logging.debug("A goal state has been reached.")
                break

    return Trajectory(states, actions)


if __name__ == '__main__':

    TRAJ_LEN_MIN = 10
    TRAJ_LEN_MAX = 10
    OPTIMAL_TRACES = 0
    MAX_PLANNING_TIME = 600
    MAX_REPLANNING_TIME = 60  # time to check problem feasibility
    MAX_RANDOM_TRIALS = 3  # maximum number of random action samplings at each step
    DOWNWARD_SEARCH_CFG = 'let(hff,ff(),let(hcea,cea(),lazy_greedy([hff,hcea],preferred=[hff,hcea])))'
    HEUR_PLANNER_CFG = {
        'name': 'fast-downward',
        'params': dict(
            fast_downward_search_config=DOWNWARD_SEARCH_CFG,
            fast_downward_search_time_limit=f"{MAX_PLANNING_TIME}s"
        )}
    OPT_PLANNER_CFG = {
        'name': 'fast-downward-opt',
    }

    PLANNER_CFG = OPT_PLANNER_CFG
    logging.basicConfig(
        filename='out.log',
        level=logging.DEBUG
    )
    GEN_DIR = "pddl-generators"
    BENCHMARK_DIR = "benchmarks"
    DOMAINS_DIR = "domains"
    PROB_DIR = "problems/learning_hard"
    TRAJ_DIR = "trajectories/learning_hard"
    DOM_CFG = f"{BENCHMARK_DIR}/problems_learning_hard.yaml"

    # Trace CPU time
    run_start = datetime.now()

    # Instantiate a PDDL problem reader
    reader = PDDLReader()

    # Disable printing of planning engine credits to avoid overloading stdout
    unified_planning.shortcuts.get_environment().credits_stream = None

    # Read domain configs
    with open(f"../{DOM_CFG}") as f:
        cfg = yaml.safe_load(f)
        seed = cfg['SEED']
        domains = cfg['domains']

    to_be_avoided = []
    domains = {k: v for k, v in domains.items() if k not in to_be_avoided}

    for domain in domains:

        # Set random seed for reproducibility
        np.random.seed(seed)
        random.seed(seed)

        # Trace CPU time
        domain_run_start = datetime.now()

        # Clean domain problems directory
        if os.path.exists(f"../{BENCHMARK_DIR}/{PROB_DIR}/{domain}"):
            shutil.rmtree(f"../{BENCHMARK_DIR}/{PROB_DIR}/{domain}")
        os.makedirs(os.path.join(f"../{BENCHMARK_DIR}/{PROB_DIR}/{domain}"))

        # Clean domain trajectories directory
        if os.path.exists(f"../{BENCHMARK_DIR}/{TRAJ_DIR}/{domain}"):
            shutil.rmtree(f"../{BENCHMARK_DIR}/{TRAJ_DIR}/{domain}")
        os.makedirs(os.path.join(f"../{BENCHMARK_DIR}/{TRAJ_DIR}/{domain}"))

        with alive_bar(len(domains[domain]),
                       title=f'Processing domain {domain}',
                       length=20,
                       bar='halloween') as bar:
            # For every domain problem kwargs
            for i, kwargs in enumerate(domains[domain]):

                if i >= OPTIMAL_TRACES:
                    PLANNER_CFG = HEUR_PLANNER_CFG
                else:
                    PLANNER_CFG = OPT_PLANNER_CFG

                trajectory = Trajectory([], [])

                while len(trajectory.states) < TRAJ_LEN_MIN:
                    # Generate a problem
                    logging.debug(f"Generating a new problem")
                    kwargs['seed'] = np.random.randint(1, 1000)
                    generate_prob = getattr(sys.modules[__name__], f'problem_{domain}')
                    problem_str = generate_prob(**kwargs)
                    # Fix hyphens to avoid issues with unified-planning parsing
                    problem_str = re.sub(r'(?<=\w)-(?=\w)', '_', problem_str)

                    # Write the problem string to a file
                    problems_dir = f"../{BENCHMARK_DIR}/{PROB_DIR}/{domain}"
                    problem_file = f'{problems_dir}/{len(os.listdir(problems_dir))}_{domain}_prob.pddl'
                    with open(problem_file, 'w') as f:
                        f.write(problem_str.lower())

                    # Parse the problem in unified-planning
                    domain_file = f'../{BENCHMARK_DIR}/{DOMAINS_DIR}/{domain}.pddl'
                    problem = reader.parse_problem(domain_file, problem_file)

                    # Generate a trace by solving the problem
                    trajectory = generate_traj(problem)
                    if len(trajectory.states) < TRAJ_LEN_MIN:
                        logging.debug(f"Failed to generate a sufficiently long trace. Retrying...")
                        os.remove(problem_file)

                trace_file = f'../{BENCHMARK_DIR}/{TRAJ_DIR}/{domain}/{len(os.listdir(f"../{BENCHMARK_DIR}/{TRAJ_DIR}/{domain}"))}_{domain}_traj'
                trajectory.write(trace_file)
                bar()  # update progress bar

        logging.info(f'{domain} CPU time (s): {(datetime.now() - domain_run_start).seconds}')

    logging.info(f'Total CPU time (s): {(datetime.now() - run_start).seconds}')





import torch
import torch.nn as nn
import torch.nn.functional as F

import itertools
import math
import string
import json
from json import JSONEncoder


class Type:
    def __init__(self, name, parent):
        self.name = name
        self.parent = parent

    def is_child(self, another_type):
        if self.name == another_type.name:
            return True
        elif self.parent is None:
            return False
        else:
            return self.parent.is_child(another_type)


class Predicate:
    def __init__(self, name, params):
        self.name = name
        # params are dicts {Type: num}
        self.params = params
        # self.params_types = sorted(params.keys(), key=lambda x: x.name)
        self.params_types = list(params.keys())


    def proposition(self, sorted_obj_lists):
        return (
                self.name
                + " "
                + " ".join(
            obj
            for obj_list in sorted_obj_lists
            for obj in obj_list
        )
        ).strip()

    def ground(self, objects):
        """
        Input a list of objects in the form {Type: []}
        Return all the propositions grounded from this predicates with the objects
        """
        propositions = []
        obj_lists_per_params = {params_type: [] for params_type in self.params_types}
        for params_type in self.params_types:
            for obj_type in objects.keys():
                if obj_type.is_child(params_type):
                    obj_lists_per_params[params_type].extend(objects[obj_type])
        for obj_lists in itertools.product(
            *[
                itertools.permutations(
                    obj_lists_per_params[params_type], self.params[params_type]
                )
                for params_type in self.params_types
            ]
        ):
            propositions.append(self.proposition(obj_lists))
        return propositions


class Action_Schema(nn.Module):
    def __init__(self, name, params):
        super(Action_Schema, self).__init__()
        self.name = name
        # params are dicts {Type: num}
        self.params = params
        self.params_types = list(params.keys())
        # predicates that are relevant
        self.predicates = []

    def action(self, sorted_obj_lists):
        return (
                self.name
                + " "
                + " ".join(
            obj
            for obj_list in sorted_obj_lists
            for obj in obj_list
        )
        ).strip()

    def initialise(self, predicates, device):
        """
        Input all predicates and generate the deep learning model for the action schema
        """
        n_features = 0
        for predicate in predicates:
            # A predicate is relevant to an action schema iff for each of its param type,
            # the number of objects required is leq the number of objects there is
            # for the same type or children type in the action schema
            is_relevant = True
            # Also calculate how many propositions there are when predicate is grounded on "variables"
            # e.g. on X Y; on Y X when X and Y are variables
            n_ground = 1
            for params_type in predicate.params_types:
                n_params = 0
                for model_params_type in self.params:
                    if model_params_type.is_child(params_type):
                        n_params += self.params[model_params_type]
                if predicate.params[params_type] > n_params:
                    is_relevant = False
                    break
                else:
                    n_ground *= math.perm(n_params, predicate.params[params_type])
            if is_relevant:
                self.predicates.append(predicate)
                n_features += n_ground
        n_features = int(n_features)

        self.randn = torch.randn(n_features, 128, device=device, requires_grad=True)
        self.mlp = nn.Sequential(
            nn.Linear(128, 64),
            nn.ReLU(),
            nn.Linear(64, 32),
            nn.ReLU(),
            nn.Linear(32, 16),
            nn.ReLU(),
            nn.Linear(16, 4),
            nn.Softmax(dim=1),
        )
        self.mlp.to(device)

    def forward(self):
        return self.mlp(self.randn)

    def ground(self, objects):
        actions = []
        propositions = []
        obj_lists_per_params = {params_type: [] for params_type in self.params_types}
        for params_type in self.params_types:
            for obj_type in objects.keys():
                if obj_type.is_child(params_type):
                    obj_lists_per_params[params_type].extend(objects[obj_type])
        for obj_list in itertools.product(
            *[
                itertools.permutations(
                    obj_lists_per_params[params_type], self.params[params_type]
                )
                for params_type in self.params_types
            ]
        ):
            actions.append(self.action(obj_list))
            objects_per_action = {}
            for i in range(len(self.params_types)):
                objects_per_action[self.params_types[i]] = obj_list[i]
            propositions_per_action = []
            for predicate in self.predicates:
                propositions_per_action.extend(predicate.ground(objects_per_action))
            propositions.append(propositions_per_action)
        return actions, propositions

    def pretty_print(self,params):
        var = {}
        for params_type in self.params_types:
            var[params_type] = params[params_type.name]
        propositions = [
            p for predicate in self.predicates for p in predicate.ground(var)
        ]
        precon_list = []
        addeff_list = []
        deleff_list = []
        result = torch.argmax(self(), dim=1)
        for i in range(len(propositions)):
            if result[i] == 1:
                addeff_list.append(propositions[i])
            elif result[i] == 2:
                precon_list.append(propositions[i])
            elif result[i] == 3:
                precon_list.append(propositions[i])
                deleff_list.append(propositions[i])
        return precon_list, addeff_list, deleff_list, var


class Domain_Model(nn.Module):
    def __init__(self, types, predicates, action_schemas, device):
        super(Domain_Model, self).__init__()
        self.types = types
        self.predicates = predicates
        self.action_schemas = action_schemas
        self.device = device
        for action_schema in action_schemas:
            action_schema.initialise(predicates, self.device)

    def ground(self, objects):
        # Ground predicates to propositions, action schemas to actions
        # Record in a dictionary with values as indices, for later lookup
        self.propositions = {}
        self.actions = {}
        for predicate in self.predicates:
            for proposition in predicate.ground(objects):
                self.propositions[proposition] = len(self.propositions)

        # For each action schema, ground to actions and then find the indices
        self.indices = []
        # Also need to know which action schema each action is from
        self.action_to_schema = []
        for action_schema in self.action_schemas:
            grounded_actions, relevant_props = action_schema.ground(objects)
            for action in grounded_actions:
                self.actions[action] = len(self.actions)
                self.action_to_schema.append(action_schema)
            for propositions in relevant_props:
                self.indices.append([self.propositions[p] for p in propositions])

    def build(self, actions):
        """
        actions is a list of numbers
        """
        precon = torch.zeros(
            (len(actions), len(self.propositions)),
            device=self.device,
            requires_grad=False,
        )
        addeff = torch.zeros(
            (len(actions), len(self.propositions)),
            device=self.device,
            requires_grad=False,
        )
        deleff = torch.zeros(
            (len(actions), len(self.propositions)),
            device=self.device,
            requires_grad=False,
        )
        for i in range(len(actions)):
            y_indices = self.indices[actions[i]]
            schema = self.action_to_schema[actions[i]]
            y_indices_set = set(y_indices)

            schema_prams = schema()
            schema_precon = schema_prams @ torch.tensor(
                [0.0, 0.0, 1.0, 1.0], device=self.device
            )
            schema_addeff = schema_prams @ torch.tensor(
                [0.0, 1.0, 0.0, 0.0], device=self.device
            )
            schema_deleff = schema_prams @ torch.tensor(
                [0.0, 0.0, 0.0, 1.0], device=self.device
            )

            if len(y_indices) > len(y_indices_set):
                # There are duplicate indices in y_indices
                # Multiple predicates are grounded to one same proposition
                # We need to combine the contribution from different predicates to one proposition
                applied = set()
                for y_idx in y_indices:
                    if y_idx not in applied:
                        precon[i, y_idx] += schema_precon[y_idx]
                        addeff[i, y_idx] += schema_addeff[y_idx]
                        deleff[i, y_idx] += schema_deleff[y_idx]
                        applied.add(y_idx)
                    else:
                        # The multiple effects are combined with "or"
                        # p v q = not ((not p)^(not q))
                        precon[i, y_idx] = 1 - (1 - precon[i, y_idx]) * (
                            1 - schema_precon[y_idx]
                        )
                        addeff[i, y_idx] = 1 - (1 - addeff[i, y_idx]) * (
                            1 - schema_addeff[y_idx]
                        )
                        deleff[i, y_idx] = 1 - (1 - deleff[i, y_idx]) * (
                            1 - schema_deleff[y_idx]
                        )
            else:
                x_indices = [i] * len(y_indices)
                precon[x_indices, y_indices] += schema_precon
                addeff[x_indices, y_indices] += schema_addeff
                deleff[x_indices, y_indices] += schema_deleff
        return precon, addeff, deleff

    @staticmethod
    def create_from_json(json_dict, device):
        type_dict = {}
        predicates = []
        action_schemas = []
        for t in json_dict["types"]:
            type_dict[t["name"]] = Type(t["name"], type_dict.get(t["parent"]))
        for p in json_dict["predicates"]:
            predicates.append(
                Predicate(
                    p["name"],
                    {type_dict[param]: num for param, num in p["params"].items()},
                )
            )
        for a in json_dict["action_schemas"]:
            action_schemas.append(
                Action_Schema(
                    a["name"],
                    {type_dict[param]: num for param, num in a["params"].items()},
                )
            )
        return Domain_Model(
            list(type_dict.values()), predicates, action_schemas, device
        )

    def ground_from_json(self, file_pth):
        with open(file_pth, "r") as f:
            json_dict = json.load(f)
        type_dict = {t.name: t for t in self.types}
        self.ground(
            {
                type_dict[type_num]: object_list
                for type_num, object_list in json_dict.items()
            }
        )
    def ground_from_dict(self, json_dict):
        type_dict = {t.name: t for t in self.types}
        self.ground(
            {
                type_dict[type_num]: object_list
                for type_num, object_list in json_dict.items()
            }
        )

    def state_dict(self):
        pam_state_dicts = {}
        for schema in self.action_schemas:
            pam_state_dicts[schema.name] = [schema.randn, schema.state_dict()]
        return pam_state_dicts

    def load_state_dict(self, pam_state_dicts):
        for schema in self.action_schemas:
            schema.randn = pam_state_dicts[schema.name][0]
            schema.load_state_dict(pam_state_dicts[schema.name][1])


class DomainModelEncoder(JSONEncoder):
    def default(self, o):
        if isinstance(o, Type):
            return {
                "name": o.name,
                "parent": o.parent.name if o.parent is not None else None,
            }
        elif isinstance(o, Predicate) or isinstance(o, Action_Schema):
            return {
                "name": o.name,
                "params": {param.name: num for param, num in o.params.items()},
            }
        elif isinstance(o, Domain_Model):
            return {
                "types": o.types,
                "predicates": o.predicates,
                "action_schemas": o.action_schemas,
            }
        else:
            return o.__dict__



def dump_model(domain_model, file_pth):
    with open(file_pth, "w") as f:
        f.write(json.dumps(domain_model, cls=DomainModelEncoder, indent=4))


def load_model(file_pth, device=torch.device("cpu")):
    with open(file_pth, "r") as f:
        domain_model = Domain_Model.create_from_json(json.load(f), device)
    return domain_model



#! /usr/bin/env python3

import copy
import math
import euclidean_graph
import sys
import random

sys.path.append("../")
sys.path.append("../../")

MAX_SEED = 10000000


def road(f, t, length):
    length = int(math.ceil(length / 10.0))
    print(f"  (road {f} {t})")
    print("  (= (road-length %s %s) %d)" % (f, t, length))


def symmetric_road(f, t, length):
    road(f, t, length)
    road(t, f, length)


def min_required_fuel(graph, nodes, size, start):
    dist = {}
    for v in graph.vertices:
        dist[v] = size * size * nodes * 4  # upper bound ...

    dist[start] = 0
    queue = copy.copy(graph.vertices)
    assert start in queue

    while len(queue) > 0:
        mindist = size * size * nodes * 4
        for x in queue:
            if dist[x] < mindist:
                mindist = dist[x]
                minelement = x
        queue.remove(minelement)
        for v, u in graph.edges:
            if u == minelement and v in queue:
                ndist = (
                    dist[minelement] + minelement.round_distance(v) * 2
                )  # *2; conversion from distance to fuel req
                if ndist < dist[v]:
                    dist[v] = ndist

    ret = 0
    for x in list(dist.keys()):
        if dist[x] > ret:
            ret = dist[x]
    return ret * 2


if len(sys.argv) != 8:
    raise SystemExit(
        "Usage: <nodes> <size^(1/2)> <degree> <mindistance> <nr-trucks> <nr-packages> <seed>"
    )

nodes = int(sys.argv[1])
size = int(sys.argv[2])
degree = float(sys.argv[3])
epsilon = float(sys.argv[4])
trucks = int(sys.argv[5])
packages = int(sys.argv[6])
seed = float(sys.argv[7])

max_capacity = 4  # maximum number of packages in one truck
assert max_capacity > 2

if not seed:
    seed = random.randrange(MAX_SEED) + 1
random.seed(seed)

#         deg * width * height
# ratio = ---------------------------
#         nodes * pi * Connect^2
connect_distance = math.sqrt((degree * size * size) / (nodes * math.pi * 0.694))

city_a = euclidean_graph.generate_connected_safe(
    nodes, size, size, connect_distance, epsilon
)
city_b = euclidean_graph.generate_connected_safe(
    nodes, size, size, connect_distance, epsilon
)


id = (
    "two-cities-sequential-%dnodes-%dsize-%ddegree-%dmindistance-%dtrucks-%dpackages-%dseed"
    % (nodes, size, degree, epsilon, trucks, packages, seed)
)

print("; Transport %s" % id)
print()

print("(define (problem transport-%s)" % id)
print(" (:domain transport)")
print(" (:objects")

for i in range(nodes):
    print("  city-1-loc-%d - location" % (i + 1))
    print("  city-2-loc-%d - location" % (i + 1))
for i in range(trucks):
    print("  truck-%d - vehicle" % (i + 1))

for i in range(packages):
    print("  package-%d - package" % (i + 1))

for i in range(max_capacity + 1):
    print("  capacity-%d - capacity-number" % i)

print(" )")
print(" (:init")

print("  (= (total-cost) 0)")


for i in range(max_capacity):
    print("  (capacity-predecessor capacity-%d capacity-%d)" % (i, i + 1))


for u, v in city_a.edges:
    print("  ; %d,%d -> %d,%d" % (u.x, u.y, v.x, v.y))

    dist = u.round_distance(v)
    road("city-1-" + u.name, "city-1-" + v.name, dist)
for u, v in city_b.edges:
    print("  ; %d,%d -> %d,%d" % (u.x + 2 * size, u.y, v.x + 2 * size, v.y))

    dist = u.round_distance(v)
    road("city-2-" + u.name, "city-2-" + v.name, dist)

# connect the two cities
min_connect_distance = 4 * size
for v in city_a.vertices:
    for u in city_b.vertices:
        if (
            v.distance(euclidean_graph.Point(u.name, u.x + 2 * size, u.y))
            < min_connect_distance
        ):
            min_tuple = (v, u)
            min_connect_distance = v.distance(
                euclidean_graph.Point(u.name, u.x + 2 * size, u.y)
            )

print(
    "  ; %d,%d <-> %d,%d"
    % (min_tuple[0].x, min_tuple[0].y, min_tuple[1].x + 2 * size, min_tuple[1].y)
)

symmetric_road(
    "city-1-" + min_tuple[0].name, "city-2-" + min_tuple[1].name, min_connect_distance
)

for i in range(packages):
    print("  (at package-%d city-1-%s)" % (i + 1, random.choice(city_a.vertices).name))

required_fuel = max(
    min_required_fuel(city_a, nodes, size, min_tuple[0]),
    min_required_fuel(city_b, nodes, size, min_tuple[1]),
    min_connect_distance / 2,
)
for i in range(trucks):
    print("  (at truck-%d city-2-%s)" % (i + 1, random.choice(city_a.vertices).name))
    capacity = random.randint(2, 4)
    print("  (capacity truck-%d capacity-%d)" % (i + 1, capacity))


print(" )")
print(" (:goal (and")

for i in range(packages):
    print("  (at package-%d city-2-%s)" % (i + 1, random.choice(city_b.vertices).name))

print(" ))")

print(" (:metric minimize (total-cost))")

print(")")



import difflib
import os
import re
import shutil
import subprocess
import numpy as np
import yaml
from unified_planning.io import PDDLReader, PDDLWriter


def remove_trajs(traj_dir: str,
                 num: int = np.inf) -> None:
    """
    Remove trajectories with an ID higher than the given one.
    :param traj_dir: directory with the set of trajectories
    :param traj_dir: maximum number of trajectories
    :return:
    """
    for d in os.listdir(traj_dir):
        for t in os.listdir(os.path.join(traj_dir, d)):
            if int(t.split('_')[0]) >= num:
                os.remove(os.path.join(traj_dir, d, t))


def print_actions_with_no_effs(traj_dir: str = '../benchmarks/trajectories') -> None:
    """
    Compare trajectories between two sets of trajectories.
    :param traj_dir: directory with the set of trajectories
    :return:
    """

    for d in os.listdir(traj_dir):

        for t in os.listdir(os.path.join(traj_dir, d)):
            with open(os.path.join(traj_dir, d, t), 'r') as f:
                traj = f.readlines()
            states = [r for r in traj if r.startswith('(:state ')]
            actions = [r for r in traj if r.startswith('(:action ')]

            for i in range(len(states) - 1):
                if states[i] == states[i + 1]:
                    print(f"Domain {d} action {actions[i]}")


def compare_trajs(traj_dir1: str = '../benchmarks/trajectories',
                  traj_dir2: str = '../benchmarks/trajectoriesbkp') -> None:
    """
    Compare trajectories between two sets of trajectories.
    :param traj_dir1: directory with the first set of trajectories
    :param traj_dir2: directory with the second set of trajectories
    """

    assert np.all(sorted(os.listdir(traj_dir1)) == sorted(os.listdir(traj_dir2)))

    for d in os.listdir(traj_dir1):
        assert np.all(sorted(os.listdir(f"{traj_dir1}/{d}")) == sorted(os.listdir(f"{traj_dir2}/{d}")))
        for t in os.listdir(os.path.join(traj_dir1, d)):
            with open(os.path.join(traj_dir1, d, t), 'r') as f:
                data1 = f.readlines()
            with open(os.path.join(traj_dir2, d, t), 'r') as f:
                data2 = f.readlines()

            for line in difflib.unified_diff(data1, data2, fromfile='file1', tofile='file2', lineterm=''):
                print(f'File: {t}')
                print(line)


def get_applicable_actions_val(problem_file, domain_file):

    tmp_domain_file = f"{domain_file}.tmp"
    tmp_problem_file = f"{problem_file}.tmp"
    shutil.copy(domain_file, tmp_domain_file)
    shutil.copy(problem_file, tmp_problem_file)

    with open(tmp_domain_file, 'r') as f:
        domain_text = f.read()

    new_text = re.sub(r":effect\s*\(and(?:\s*\(.*\))*", ":effect (and )", domain_text)

    with open(tmp_domain_file, 'w') as f:
        f.write(new_text)

    # Ground the problem using Val
    bash_command = ["./val/Instantiate", tmp_domain_file, tmp_problem_file]
    process = subprocess.run(bash_command, capture_output=True)
    output = str(process.stdout).split('\\n')

    os.remove(tmp_domain_file)
    os.remove(tmp_problem_file)

    all_actions = re.findall(r"\([^()]*\)", re.findall("so far.*literals", "".join(output))[0])
    all_actions = [a[1:-1].split() for a in all_actions]
    all_actions = {f"({a[0].rsplit('_', 1)[0]} {' '.join(a[1:])})" for a in all_actions}

    return all_actions


# Replace "word-word" with "word_word", otherwise unified-planning does not correctly write the problem file
def fix_domain_format(dom_dir: str = '../benchmarks/domains') -> None:
    """
    Replace "word-word" with "word_word", otherwise unified-planning does not correctly write the problem file
    :param dom_dir: directory of domains
    :return:
    """

    for d in os.listdir(dom_dir):
        with open(f"{dom_dir}/{d}", 'r') as f:
            data = f.read()
        with open(f"{dom_dir}/{d}", 'w') as f:

            # Changhe hyphens for unified-planning
            data = re.sub(r'(?<=\w)-(?=\w)', '_', data)

            # Ferry: previous step change not-eq into not_eq --> remove not_ for OffLAM
            # --> TODO 1: open issue in OffLAM
            # --> TODO 2: open issue in unified-planning
            data = data.replace('(not_eq ', '(noteq ')

            # lower format
            # TODO: check goldminer without lower format
            data = data.lower()

            f.write(data)


def preprocess_trace(traj_path: str) -> None:

    for d in os.listdir(traj_path):
        dom_traj = dict()
        for t in os.listdir(f"{traj_path}/{d}"):
            with open(f"{traj_path}/{d}/{t}", 'r') as f:
                traj_str = f.readlines()
            states = [s for s in traj_str if s.startswith('(:state')]
            dom_traj[f"{traj_path}/{d}/{t}"] = len(states)


def reduce_problem_settings(in_problem_path: str,
                            out_problem_path: str) -> None:
    with open(in_problem_path, 'r') as f:
        settings = yaml.safe_load(f)

    for d in settings['domains']:
        settings['domains'][d] = settings['domains'][d][-2:]

    with open(out_problem_path, 'w') as f:
        yaml.dump(settings, f)


def empty_domain(domain_path: str, empty_domain_path: str = 'empty.pddl'):

    pddl_domain = PDDLReader().parse_problem(domain_path)

    # Loop through all actions and remove preconditions and effects
    for action in pddl_domain.actions:
        action.clear_preconditions()
        action.clear_effects()

    domain_str = PDDLWriter(pddl_domain).get_domain()
    pattern = re.compile(
        r"(:action[\s\S]*?:parameters\s*\([^)]*\))\)",
        re.MULTILINE
    )

    replacement = r"\1\n  :precondition (and )\n  :effect (and ))\n"

    # TODO: open issue in up
    domain_str = domain_str.replace(f"(domain {pddl_domain.name}-domain)",
                                    f"(domain {pddl_domain.name})")

    with open(empty_domain_path, 'w') as f:
        f.write(pattern.sub(replacement, domain_str))

    return empty_domain_path


if __name__ == '__main__':
    pass

    # remove_trajs('../benchmarks/problems/applicability', 100)
    # remove_trajs('../benchmarks/trajectories/applicability', 100)

    # LIMIT DATASETS TO 100 MAX TRACES
    # remove_trajs('../benchmarks/problems/predictive_power', 100)
    # states_path = '../benchmarks/states/predictive_power'
    # for domain in os.listdir(states_path):
    #     domain_states_path = f"{states_path}/{domain}"
    #     with open(f"{domain_states_path}/test_states.json", 'r') as f:
    #         test_set = json.load(f)
    #     new_test_set = {k:v for k, v in test_set.items() if int(k.split('_')[0]) < 100}
    #     with open(f"{domain_states_path}/test_states.json", 'w') as f:
    #         json.dump(new_test_set, f, indent=3)

    # preprocess_trace('../benchmarks/trajectories/learning')

    # reduce_problem_settings('../benchmarks/problems_learning.yaml',
    #                         '../benchmarks/problems_learning_hard.yaml')



import sys
from collections import defaultdict
import numpy

numpy.random.seed(18)
ipc_instances = sys.argv[1:]

instances_by_key = defaultdict(set)
for instance in ipc_instances:
    (_, _, plan_len, total_size, dimensions, target) = instance.split("-")
    (_, _, _, _, num_towers, seed) = target.split("_")

    instances_by_key[(dimensions, num_towers)].add(instance)

(size_x, size_y) = (4, 3)
for range_num_towers, range_height in [
    ([1, 2], range(3, 8)),
    ([3, 4], range(3, 6)),
    ([5, 6], range(3, 5)),
]:
    for height in range_height:
        for num_towers in range_num_towers:
            dimensions = f"{size_x}x{size_y}x{height}"
            instance_set = sorted(instances_by_key[(dimensions, str(num_towers))])
            random_selected_instance = numpy.random.choice(instance_set)
            print(random_selected_instance)



#! /usr/bin/env python3

import math
import euclidean_graph
import sys
import random

sys.path.append("../")
sys.path.append("../../")

MAX_SEED = 10000000


def road(f, t, length):
    length = int(math.ceil(length / 10.0))
    print(f"  (road {f} {t})")
    print("  (= (road-length %s %s) %d)" % (f, t, length))


def symmetric_road(f, t, length):
    road(f, t, length)
    road(t, f, length)


def shortest_route(ca, cb, ox, oy):
    # connect the two cities
    min_connect_distance = -1.0
    for v in city_a.vertices:
        for u in city_b.vertices:
            if (
                v.distance(euclidean_graph.Point(u.name, u.x + ox, u.y + oy))
                < min_connect_distance
                or min_connect_distance == -1.0
            ):
                min_tuple = (v, u)
                min_connect_distance = v.distance(
                    euclidean_graph.Point(u.name, u.x + 2 * size, u.y)
                )

    return min_tuple, min_connect_distance


if len(sys.argv) != 8:
    raise SystemExit(
        "Usage: <nodes> <size^(1/2)> <degree> <mindistance> <nr-trucks> <nr-packages> <seed>"
    )

nodes = int(sys.argv[1])
size = int(sys.argv[2])
degree = float(sys.argv[3])
epsilon = float(sys.argv[4])
trucks = int(sys.argv[5])
packages = int(sys.argv[6])
seed = float(sys.argv[7])

max_capacity = 4  # maximum number of packages in one truck
assert max_capacity > 2

if not seed:
    seed = random.randrange(MAX_SEED) + 1
random.seed(seed)

#         deg * width * height
# ratio = ---------------------------
#         nodes * pi * Connect^2
connect_distance = math.sqrt((degree * size * size) / (nodes * math.pi * 0.694))

city_a = euclidean_graph.generate_connected_safe(
    nodes, size, size, connect_distance, epsilon
)
city_b = euclidean_graph.generate_connected_safe(
    nodes, size, size, connect_distance, epsilon
)
city_c = euclidean_graph.generate_connected_safe(
    nodes, size, size, connect_distance, epsilon
)


id = (
    "three-cities-sequential-%dnodes-%dsize-%ddegree-%dmindistance-%dtrucks-%dpackages-%dseed"
    % (nodes, size, degree, epsilon, trucks, packages, seed)
)

print("; Transport %s" % id)
print()

print("(define (problem transport-%s)" % id)
print(" (:domain transport)")
print(" (:objects")

for i in range(nodes):
    print("  city-1-loc-%d - location" % (i + 1))
    print("  city-2-loc-%d - location" % (i + 1))
    print("  city-3-loc-%d - location" % (i + 1))
for i in range(trucks):
    print("  truck-%d - vehicle" % (i + 1))

for i in range(packages):
    print("  package-%d - package" % (i + 1))

for i in range(max_capacity + 1):
    print("  capacity-%d - capacity-number" % i)

print(" )")
print(" (:init")

print("  (= (total-cost) 0)")


for i in range(max_capacity):
    print("  (capacity-predecessor capacity-%d capacity-%d)" % (i, i + 1))

for u, v in city_a.edges:
    print("  ; %d,%d -> %d,%d" % (u.x, u.y, v.x, v.y))

    dist = u.round_distance(v)
    road("city-1-" + u.name, "city-1-" + v.name, dist)
for u, v in city_b.edges:
    print("  ; %d,%d -> %d,%d" % (u.x + 2 * size, u.y, v.x + 2 * size, v.y))

    dist = u.round_distance(v)
    road("city-2-" + u.name, "city-2-" + v.name, dist)
for u, v in city_c.edges:
    print(
        "  ; %d,%d -> %d,%d" % (u.x + size, u.y + 2 * size, v.x + size, v.y + 2 * size)
    )

    dist = u.round_distance(v)
    road("city-3-" + u.name, "city-3-" + v.name, dist)

connect_ab, dist_ab = shortest_route(city_a, city_b, 2 * size, 0)
print(
    "  ; %d,%d <-> %d,%d"
    % (connect_ab[0].x, connect_ab[0].y, connect_ab[1].x + 2 * size, connect_ab[1].y)
)
symmetric_road("city-1-" + connect_ab[0].name, "city-2-" + connect_ab[1].name, dist_ab)
connect_ac, dist_ac = shortest_route(city_a, city_c, size, 2 * size)
symmetric_road("city-1-" + connect_ac[0].name, "city-3-" + connect_ac[1].name, dist_ac)
connect_bc, dist_bc = shortest_route(city_b, city_c, size, -2 * size)
symmetric_road("city-2-" + connect_bc[0].name, "city-3-" + connect_bc[1].name, dist_bc)

package_loc = {}
for i in range(packages):
    package_loc["package-%d" % (i + 1)] = (
        random.randint(1, 3),
        random.choice(city_a.vertices).name,
    )
    print(
        "  (at package-%d city-%d-%s)"
        % (
            i + 1,
            package_loc["package-%d" % (i + 1)][0],
            package_loc["package-%d" % (i + 1)][1],
        )
    )

for i in range(trucks):
    print(
        "  (at truck-%d city-%d-%s)"
        % (i + 1, random.randint(1, 3), random.choice(city_a.vertices).name)
    )
    capacity = random.randint(2, 4)
    print("  (capacity truck-%d capacity-%d)" % (i + 1, capacity))


print(" )")
print(" (:goal (and")

for i in range(packages):
    rc = random.randint(1, 3)
    rl = random.choice(city_b.vertices).name
    while (
        rc == package_loc["package-%d" % (i + 1)][0]
        and rl == package_loc["package-%d" % (i + 1)][1]
    ):
        rc = random.randint(1, 3)
        rl = random.choice(city_b.vertices).name
    print("  (at package-%d city-%d-%s)" % (i + 1, rc, rl))

print(" ))")

print(" (:metric minimize (total-cost))")

print(")")
